"""Main module for the Texventure game engine."""

import argparse
import cmd
import json
import os
import sys

from rich.console import Console

from .engine import GameEngine
from .ui import UIRenderer


class TexventureShell(cmd.Cmd):
    def __init__(
        self,
        story_file=None,
        api_key=None,
        model="gpt-4.1-nano",
        use_llm=True,
        max_tokens=800,
    ):
        super().__init__()
        self.console = Console()
        self.engine = GameEngine(use_llm=use_llm, model=model, max_tokens=max_tokens)
        self.ui = UIRenderer(self.console)
        self.prompt = "(texventure) "

        if api_key:
            os.environ["OPENAI_API_KEY"] = api_key

        if story_file:
            self.load_story(story_file)

    def load_story(self, story_file):
        """Load story data from JSON file."""
        try:
            story_data = self._load_json_file(story_file)
            self.engine.load_story(story_data)
            self.prompt = f"({self.engine.story_data.title}) "

            if story_data.get("is_saved_game", False):
                self._handle_saved_game(story_data)
            else:
                self._handle_new_story()

        except (FileNotFoundError, json.JSONDecodeError) as e:
            error_msg = self._get_file_error_message(e, story_file)
            self.ui.render_error(error_msg)
            sys.exit(1)
        except Exception as e:
            self.ui.render_error(f"Error loading story file: {e}")
            sys.exit(1)

    def _load_json_file(self, story_file):
        """Load and parse JSON file."""
        with open(story_file, "r") as f:
            return json.load(f)

    def _get_file_error_message(self, error, story_file):
        """Get appropriate error message for file loading errors."""
        if isinstance(error, FileNotFoundError):
            return f"Story file '{story_file}' not found."
        elif isinstance(error, json.JSONDecodeError):
            return f"Invalid JSON in story file '{story_file}': {error}"
        return str(error)

    def _handle_saved_game(self, story_data):
        """Handle loading a saved game."""
        self.ui.render_title_screen(self.engine.story_data, "")
        self.console.print("\n[bold green]📁 Loading saved game...[/bold green]")

        event_count = len(self.engine.event_history)
        action_count = len(self.engine.action_log)
        save_point = story_data.get("saved_at", 0)

        self.console.print(
            f"[cyan]Restored {event_count} events and {action_count} actions from save point {save_point}[/cyan]"
        )

        if self.engine.game_state.current_event:
            self.console.print("\n[bold yellow]📍 Resuming from:[/bold yellow]")
            self.ui.render_event(self.engine.game_state.current_event)
        else:
            self.console.print(
                "\n[yellow]No current event found, starting fresh...[/yellow]"
            )
            first_event = self.engine.update_state("begin")
            self.ui.render_event(first_event)

    def _handle_new_story(self):
        """Handle loading a new story."""
        first_scene_desc = self._get_first_scene_description()
        self.ui.render_title_screen(self.engine.story_data, first_scene_desc)

        first_event = self.engine.initial_event()
        self.ui.render_event(first_event)

    def _get_first_scene_description(self):
        """Extract the first scene description for the title screen."""
        if not self.engine.story_data.acts:
            return ""

        first_act = self.engine.story_data.acts[0]
        act_scenes = first_act.get("scenes", [])

        if act_scenes:
            return act_scenes[0].get("description", "")
        return first_act.get("description", "")

    def do_look(self, arg):
        """Look around the current event."""
        current_event = self.engine.game_state.current_event
        if current_event:
            self.ui.render_event(current_event)
        else:
            self.ui.render_error("No current event to display.")

    def do_inventory(self, arg):
        """Show your inventory."""
        items_data = (
            getattr(self.engine.story_data, "items", {})
            if self.engine.story_data
            else {}
        )
        self.ui.render_inventory(self.engine.game_state.inventory, items_data)

    def do_take(self, item_name):
        """Take an item from the current event."""
        if not item_name.strip():
            self.ui.render_error("Please specify an item to take.")
            return

        success, message = self.engine.add_to_inventory(item_name)
        self.ui.render_success(message) if success else self.ui.render_error(message)

    def do_choice(self, choice_num):
        """Make a choice: choice 1."""
        if not choice_num.isdigit():
            self.ui.render_error("Please enter a valid choice number.")
            return

        choice_idx = int(choice_num) - 1
        success, message, new_event = self.engine.make_choice(choice_idx)

        if success:
            # Remove "You chose: " prefix if present
            feedback = message.split(": ", 1)[1] if ": " in message else message
            self.ui.render_choice_feedback(feedback)
            self.ui.render_event(new_event)
        else:
            self.ui.render_error(message)

    def do_save(self, filename):
        """Save the current game state."""
        if not filename.strip():
            title = (
                getattr(self.engine.story_data, "title", "Text Adventure")
                if self.engine.story_data
                else "Text Adventure"
            )
            filename = f"{title}-savegame.json"

        success, message = self.engine.save_game(filename)
        if success:
            self.ui.render_success(message)
            event_count = len(self.engine.event_history)
            action_count = len(self.engine.action_log)
            self.ui.render_info(
                f"Save includes {event_count} events visited and {action_count} actions taken."
            )
        else:
            self.ui.render_error(message)

    def do_history(self, arg):
        """Show the history of events visited and actions taken."""
        self.ui.render_action_history(self.engine.action_log)

    def do_status(self, arg):
        """Show current game status and statistics."""
        self.ui.render_status(
            self.engine.game_state, self.engine.event_history, self.engine.story_data
        )
        self._render_llm_status()

    def _render_llm_status(self):
        """Render LLM configuration status."""
        self.console.print("\n[bold blue]🤖 LLM Configuration[/bold blue]")
        if self.engine.use_llm:
            api_key_status = "✓ Set" if os.getenv("OPENAI_API_KEY") else "✗ Not set"
            self.console.print(f"LLM Enabled: [green]Yes[/green]")
            self.console.print(f"API Key: {api_key_status}")
            self.console.print(f"Model: {self.engine.model}")
            self.console.print(f"Max Tokens: {self.engine.max_tokens}")
        else:
            self.console.print("LLM Enabled: [red]No[/red] (using template events)")

    def default(self, line):
        """Handle numeric input as choices."""
        if line.isdigit():
            self.do_choice(line)
        else:
            self.ui.render_error(f"Unknown command: {line}")
            self.ui.render_info("Type 'help' for available commands.")

    def do_exit(self, arg):
        """Exit the game."""
        self.ui.render_goodbye()
        return True


def main():
    """Main entry point for the texventure command."""
    args = _parse_arguments()
    console = Console()

    # Validate story file exists
    if not os.path.exists(args.story_file):
        _print_file_not_found_error(console, args.story_file)
        sys.exit(1)

    # Configure LLM settings
    api_key = args.api_key or os.getenv("OPENAI_API_KEY")
    use_llm = not args.no_llm

    # Warn about missing API key
    if use_llm and not api_key:
        _print_api_key_warning(console)

    # Create and run the game
    game = TexventureShell(
        story_file=args.story_file,
        api_key=api_key,
        model=args.model,
        use_llm=use_llm,
        max_tokens=args.max_tokens,
    )

    try:
        game.cmdloop()
    except KeyboardInterrupt:
        console.print("\n[yellow]Goodbye![/yellow]")


def _parse_arguments():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(
        description="Run a text-based adventure game from a story JSON file"
    )
    parser.add_argument("story_file", help="Path to the story JSON file")
    parser.add_argument(
        "--api-key", help="OpenAI API key (or set OPENAI_API_KEY environment variable)"
    )
    parser.add_argument(
        "--model",
        help="LLM model to use for dynamic event generation",
        default="gpt-4.1-nano",
    )
    parser.add_argument(
        "--max-tokens",
        type=int,
        help="Maximum number of tokens for LLM responses (default: 800)",
        default=800,
    )
    parser.add_argument(
        "--no-llm",
        action="store_true",
        help="Disable LLM-powered dynamic event generation",
    )
    return parser.parse_args()


def _print_file_not_found_error(console, story_file):
    """Print error message when story file is not found."""
    console.print(
        f"[bold red]Error:[/bold red] Story file '[cyan]{story_file}[/cyan]' not found."
    )
    console.print("[yellow]Usage:[/yellow] [cyan]texventure <story_file.json>[/cyan]")


def _print_api_key_warning(console):
    """Print warning when LLM is enabled but no API key is provided."""
    console.print(
        "[yellow]Warning:[/yellow] LLM is enabled but no API key provided. "
        "Dynamic event generation will fall back to basic templates."
    )
    console.print(
        "[italic]Set OPENAI_API_KEY environment variable or use --api-key flag for full LLM features.[/italic]"
    )


if __name__ == "__main__":
    main()



"""Tests for the story generator module."""

import json
from unittest.mock import Mock, patch

import pytest

from src.texventure.generator import (_fill_fields_individually,
                                      collect_empty_fields, form_bulk_prompt,
                                      format_story, generate_story, llm_call,
                                      set_nested_value)


class TestCollectEmptyFields:
    """Test the collect_empty_fields function."""

    def test_simple_dict(self):
        """Test collecting empty fields from a simple dictionary."""
        data = {"title": "", "setting": "fantasy world", "outline": ""}

        empty_fields = collect_empty_fields(data)
        assert "title" in empty_fields
        assert "outline" in empty_fields
        assert "setting" not in empty_fields
        assert len(empty_fields) == 2

    def test_nested_dict(self):
        """Test collecting empty fields from nested dictionaries."""
        data = {
            "title": "Test Story",
            "acts": {"act1": {"name": "", "description": "First act"}},
            "npcs": {"hero": {"name": "Hero", "description": ""}},
        }

        empty_fields = collect_empty_fields(data)
        assert "acts.act1.name" in empty_fields
        assert "npcs.hero.description" in empty_fields
        assert len(empty_fields) == 2

    def test_list_with_dicts(self):
        """Test collecting empty fields from lists containing dictionaries."""
        data = {
            "acts": [
                {"name": "Act 1", "description": ""},
                {"name": "", "description": "Second act"},
            ]
        }

        empty_fields = collect_empty_fields(data)
        assert "acts[0].description" in empty_fields
        assert "acts[1].name" in empty_fields
        assert len(empty_fields) == 2

    def test_no_empty_fields(self):
        """Test with data that has no empty fields."""
        data = {
            "title": "Complete Story",
            "setting": "Fantasy world",
            "acts": [{"name": "Act 1", "description": "First act"}],
        }

        empty_fields = collect_empty_fields(data)
        assert len(empty_fields) == 0


class TestSetNestedValue:
    """Test the set_nested_value function."""

    def test_simple_path(self):
        """Test setting a value with a simple path."""
        data = {"title": "", "setting": ""}
        set_nested_value(data, "title", "New Title")
        assert data["title"] == "New Title"
        assert data["setting"] == ""

    def test_nested_path(self):
        """Test setting a value with a nested path."""
        data = {"acts": {"act1": {"name": "", "description": ""}}}
        set_nested_value(data, "acts.act1.name", "Chapter One")
        assert data["acts"]["act1"]["name"] == "Chapter One"

    def test_array_path(self):
        """Test setting a value in an array."""
        data = {
            "acts": [{"name": "", "description": ""}, {"name": "", "description": ""}]
        }
        set_nested_value(data, "acts[1].name", "Second Act")
        assert data["acts"][1]["name"] == "Second Act"
        assert data["acts"][0]["name"] == ""


class TestFormBulkPrompt:
    """Test the form_bulk_prompt function."""

    def test_single_field(self):
        """Test prompt formation with a single field."""
        fields = ["title"]
        context = '{"title": "", "setting": "fantasy"}'

        prompt = form_bulk_prompt(fields, context)
        assert "title" in prompt
        assert "JSON object" in prompt
        assert context in prompt

    def test_multiple_fields(self):
        """Test prompt formation with multiple fields."""
        fields = ["title", "setting", "outline"]
        context = '{"title": "", "setting": "", "outline": ""}'

        prompt = form_bulk_prompt(fields, context)
        for field in fields:
            assert field in prompt
        assert "JSON object" in prompt


class TestLLMCall:
    """Test the llm_call function."""

    @patch("src.texventure.generator.requests.post")
    def test_successful_call(self, mock_post):
        """Test a successful LLM API call."""
        mock_response = Mock()
        mock_response.json.return_value = {
            "choices": [{"message": {"content": "Generated content"}}]
        }
        mock_response.raise_for_status.return_value = None
        mock_post.return_value = mock_response

        result = llm_call("test prompt", "test-api-key")
        assert result == "Generated content"

        # Verify the request was made correctly
        mock_post.assert_called_once()
        args, kwargs = mock_post.call_args
        assert "json" in kwargs
        assert kwargs["json"]["messages"][1]["content"] == "test prompt"

    @patch("src.texventure.generator.requests.post")
    def test_json_mode(self, mock_post):
        """Test LLM call with JSON mode enabled."""
        mock_response = Mock()
        mock_response.json.return_value = {
            "choices": [{"message": {"content": '{"title": "Test"}'}}]
        }
        mock_response.raise_for_status.return_value = None
        mock_post.return_value = mock_response

        result = llm_call("test prompt", "test-api-key", force_json=True)
        assert result == '{"title": "Test"}'

        # Verify JSON mode was enabled
        args, kwargs = mock_post.call_args
        assert kwargs["json"]["response_format"] == {"type": "json_object"}
        assert "valid JSON" in kwargs["json"]["messages"][0]["content"]

    @patch("src.texventure.generator.requests.post")
    def test_api_error(self, mock_post):
        """Test handling of API errors."""
        mock_post.side_effect = Exception("API Error")

        result = llm_call("test prompt", "test-api-key")
        assert result is None


class TestGenerateStory:
    """Test the generate_story function."""

    @patch("src.texventure.generator.llm_call")
    def test_no_api_key(self, mock_llm_call):
        """Test behavior when no API key is provided."""
        story_data = {"title": "", "setting": ""}

        generate_story(story_data, None)

        # Should not make any LLM calls
        mock_llm_call.assert_not_called()
        # Data should remain unchanged
        assert story_data["title"] == ""
        assert story_data["setting"] == ""

    @patch("src.texventure.generator.llm_call")
    def test_no_empty_fields(self, mock_llm_call):
        """Test behavior when there are no empty fields."""
        story_data = {"title": "Complete", "setting": "Fantasy"}

        generate_story(story_data, "test-api-key")

        # Should not make any LLM calls
        mock_llm_call.assert_not_called()

    @patch("src.texventure.generator.llm_call")
    def test_successful_generation(self, mock_llm_call):
        """Test successful story generation with JSON response."""
        story_data = {"title": "", "setting": ""}

        # Mock LLM response
        mock_llm_call.return_value = (
            '{"title": "Generated Title", "setting": "Generated Setting"}'
        )

        generate_story(story_data, "test-api-key")

        # Verify LLM was called with JSON mode
        mock_llm_call.assert_called_once()
        args, kwargs = mock_llm_call.call_args
        assert kwargs.get("force_json") is True

        # Verify data was updated
        assert story_data["title"] == "Generated Title"
        assert story_data["setting"] == "Generated Setting"

    @patch("src.texventure.generator.llm_call")
    @patch("src.texventure.generator._fill_fields_individually")
    def test_json_parse_failure_fallback(self, mock_fallback, mock_llm_call):
        """Test fallback to individual filling when JSON parsing fails."""
        story_data = {"title": "", "setting": ""}

        # Mock invalid JSON response
        mock_llm_call.return_value = "Invalid JSON response"

        generate_story(story_data, "test-api-key")

        # Should fall back to individual field filling
        mock_fallback.assert_called_once()

    @patch("src.texventure.generator.llm_call")
    def test_batch_processing(self, mock_llm_call):
        """Test that large numbers of fields are processed in batches."""
        # Create story data with more than 10 empty fields
        story_data = {f"field_{i}": "" for i in range(15)}

        mock_llm_call.return_value = '{"field_0": "content"}'

        generate_story(story_data, "test-api-key")

        # Should make multiple calls due to batching
        assert mock_llm_call.call_count >= 2


class TestFillFieldsIndividually:
    """Test the _fill_fields_individually fallback function."""

    @patch("src.texventure.generator.llm_call")
    def test_individual_filling(self, mock_llm_call):
        """Test filling fields individually."""
        story_data = {"title": "", "setting": ""}
        fields = ["title", "setting"]

        # Mock individual responses
        mock_llm_call.side_effect = ["Generated Title", "Generated Setting"]

        _fill_fields_individually(story_data, fields, "test-api-key", "gpt-3.5-turbo")

        # Should make one call per field
        assert mock_llm_call.call_count == 2

        # Verify data was updated
        assert story_data["title"] == "Generated Title"
        assert story_data["setting"] == "Generated Setting"

    @patch("src.texventure.generator.llm_call")
    def test_empty_response_handling(self, mock_llm_call):
        """Test handling of empty responses."""
        story_data = {"title": ""}
        fields = ["title"]

        # Mock empty response
        mock_llm_call.return_value = ""

        _fill_fields_individually(story_data, fields, "test-api-key", "gpt-3.5-turbo")

        # Data should remain unchanged
        assert story_data["title"] == ""


class TestFormatStory:
    """Test the format_story function."""

    def test_basic_formatting(self):
        """Test basic story formatting."""
        template_data = {
            "title": "Test Story",
            "setting": "Fantasy world",
            "n_acts": 2,
            "n_scenes": 2,
            "n_npcs": 1,
        }

        result = format_story(template_data)

        assert result["title"] == "Test Story"
        assert len(result["acts"]) == 2
        assert len(result["acts"][0]["scenes"]) == 2
        assert len(result["npcs"]) == 1

    def test_creates_missing_acts(self):
        """Test that missing acts are created."""
        template_data = {
            "title": "Test",
            "n_acts": 3,
            "n_scenes": 1,
            "n_npcs": 1,
            "acts": [],
        }

        result = format_story(template_data)

        assert len(result["acts"]) == 3
        assert result["acts"][0]["name"] == "Act 1"
        assert result["acts"][1]["name"] == "Act 2"
        assert result["acts"][2]["name"] == "Act 3"

    def test_creates_missing_scenes(self):
        """Test that missing scenes are created in acts."""
        template_data = {
            "title": "Test",
            "n_acts": 1,
            "n_scenes": 3,
            "n_npcs": 1,
            "acts": [{"name": "Act 1", "description": "", "scenes": []}],
        }

        result = format_story(template_data)

        assert len(result["acts"][0]["scenes"]) == 3
        assert result["acts"][0]["scenes"][0]["name"] == "Scene 1"
        assert result["acts"][0]["scenes"][1]["name"] == "Scene 2"
        assert result["acts"][0]["scenes"][2]["name"] == "Scene 3"

    def test_creates_missing_npcs(self):
        """Test that missing NPCs are created."""
        template_data = {
            "title": "Test",
            "n_acts": 1,
            "n_scenes": 1,
            "n_npcs": 2,
            "npcs": [],
        }

        result = format_story(template_data)

        assert len(result["npcs"]) == 2
        assert result["npcs"][0]["name"] == "Character 1"
        assert result["npcs"][1]["name"] == "Character 2"



"""Fixtures and utilities for testing."""

import json
from unittest.mock import Mock

import pytest

from src.texventure.engine import GameEngine
from src.texventure.event_generators import (LLMEventGenerator,
                                             TemplateEventGenerator)
from src.texventure.models import (ActionLogEntry, Choice, Event, GameState,
                                   StoryData)
from src.texventure.save_manager import SaveLoadManager


@pytest.fixture
def sample_story_data():
    """Sample story data for testing."""
    return {
        "title": "Test Adventure",
        "setting": "A test world",
        "outline": "A simple test story",
        "acts": [
            {
                "name": "Act 1",
                "description": "The beginning",
                "scenes": [
                    {
                        "name": "Scene 1",
                        "description": "The starting scene",
                        "location": "Starting location",
                    },
                    {
                        "name": "Scene 2",
                        "description": "The second scene",
                        "location": "Second location",
                    },
                ],
            },
            {
                "name": "Act 2",
                "description": "The middle",
                "scenes": [
                    {
                        "name": "Scene 3",
                        "description": "The third scene",
                        "location": "Third location",
                    }
                ],
            },
        ],
        "npcs": [
            {"name": "Test NPC", "description": "A test character", "role": "helper"}
        ],
        "items": {
            "sword": {"name": "Magic Sword", "description": "A gleaming sword"},
            "potion": {"name": "Health Potion", "description": "Restores health"},
        },
    }


@pytest.fixture
def sample_saved_game_data(sample_story_data):
    """Sample saved game data for testing."""
    saved_data = sample_story_data.copy()
    saved_data.update(
        {
            "is_saved_game": True,
            "saved_at": 5,
            "event_counter": 5,
            "current_event": {
                "id": "event_4",
                "description": "You are in a saved location",
                "choices": [
                    {"text": "Continue forward"},
                    {"text": "Look around"},
                    {"text": "Rest"},
                    {"text": "Check inventory"},
                ],
                "items": ["sword"],
                "location": "Saved location",
                "act": 0,
                "scene": 1,
            },
            "event_history": [
                {
                    "id": "event_0",
                    "description": "The beginning",
                    "choices": [{"text": "Start"}, {"text": "Wait"}],
                    "items": [],
                    "location": "Start",
                    "act": 0,
                    "scene": 0,
                },
                {
                    "id": "event_1",
                    "description": "You started the adventure",
                    "choices": [{"text": "Go north"}, {"text": "Go south"}],
                    "items": ["potion"],
                    "location": "Crossroads",
                    "act": 0,
                    "scene": 0,
                },
            ],
            "action_log": [
                {
                    "timestamp": 0,
                    "action_type": "make_choice",
                    "details": "Made choice: Start",
                },
                {
                    "timestamp": 1,
                    "action_type": "update_state",
                    "details": "Moved to event: event_1",
                },
            ],
            "game_state": {
                "inventory": ["potion"],
                "act": 0,
                "scene": 1,
                "visited_events": ["event_0", "event_1"],
                "flags": {"started": True},
            },
        }
    )
    return saved_data


@pytest.fixture
def sample_choice():
    """Sample choice for testing."""
    return Choice("Test choice")


@pytest.fixture
def sample_event():
    """Sample event for testing."""
    choices = [
        Choice("Option 1"),
        Choice("Option 2"),
        Choice("Option 3"),
        Choice("Option 4"),
    ]
    return Event(
        id="test_event",
        description="A test event",
        choices=choices,
        items=["sword"],
        location="Test location",
        act=0,
        scene=0,
    )


@pytest.fixture
def sample_game_state():
    """Sample game state for testing."""
    return GameState(
        inventory=["potion"],
        act=0,
        scene=0,
        visited_events=["event_0"],
        flags={"test": True},
    )


@pytest.fixture
def sample_action_log_entry():
    """Sample action log entry for testing."""
    return ActionLogEntry(
        timestamp=1, action_type="test_action", details="Test action details"
    )


@pytest.fixture
def mock_llm_response():
    """Mock LLM response for testing."""
    return json.dumps(
        {
            "description": "A mysterious forest path winds before you, shrouded in mist.",
            "location": "Misty Forest Path",
            "choices": [
                {"text": "Follow the path deeper into the forest"},
                {"text": "Examine the strange markings on nearby trees"},
                {"text": "Call out to see if anyone responds"},
                {"text": "Turn back to safety"},
            ],
        }
    )


@pytest.fixture
def game_engine():
    """Game engine instance for testing."""
    return GameEngine(use_llm=False)


@pytest.fixture
def game_engine_with_llm():
    """Game engine instance with LLM enabled for testing."""
    return GameEngine(use_llm=True)


@pytest.fixture
def template_event_generator():
    """Template event generator for testing."""
    return TemplateEventGenerator()


@pytest.fixture
def mock_llm_event_generator():
    """Mock LLM event generator for testing."""
    generator = Mock(spec=LLMEventGenerator)
    return generator


@pytest.fixture
def save_manager():
    """Save manager instance for testing."""
    return SaveLoadManager()



"""Tests for UI renderer."""

from io import StringIO

import pytest
from rich.console import Console

from src.texventure.models import (ActionLogEntry, Choice, Event, GameState,
                                   StoryData)
from src.texventure.ui import UIRenderer


class TestUIRenderer:
    """Test cases for UIRenderer."""

    def test_ui_renderer_initialization_default(self):
        """Test default UI renderer initialization."""
        ui = UIRenderer()
        assert ui.console is not None
        assert isinstance(ui.console, Console)

    def test_ui_renderer_initialization_custom_console(self):
        """Test UI renderer initialization with custom console."""
        custom_console = Console(file=StringIO())
        ui = UIRenderer(custom_console)
        assert ui.console == custom_console

    def test_render_title_screen(self):
        """Test rendering title screen."""
        output = StringIO()
        console = Console(file=output, force_terminal=False, width=80)
        ui = UIRenderer(console)

        story_data = StoryData(title="Test Adventure")
        ui.render_title_screen(story_data, "A mysterious beginning...")

        rendered = output.getvalue()
        assert "Test Adventure" in rendered
        assert "TEXVENTURE" in rendered
        assert "A mysterious beginning..." in rendered

    def test_render_event(self, sample_event):
        """Test rendering an event."""
        output = StringIO()
        console = Console(file=output, force_terminal=False, width=80)
        ui = UIRenderer(console)

        ui.render_event(sample_event)

        rendered = output.getvalue()
        assert sample_event.description in rendered
        assert "Event" in rendered
        assert "What do you want to do?" in rendered

        # Check that choices are rendered
        for i, choice in enumerate(sample_event.choices, 1):
            assert f"{i}." in rendered
            assert choice.text in rendered

    def test_render_choices_with_choices(self):
        """Test rendering choices when choices are available."""
        output = StringIO()
        console = Console(file=output, force_terminal=False, width=80)
        ui = UIRenderer(console)

        choices = [
            Choice("Go north"),
            Choice("Go south"),
            Choice("Rest"),
            Choice("Check inventory"),
        ]

        ui.render_choices(choices)

        rendered = output.getvalue()
        assert "What do you want to do?" in rendered
        assert "1." in rendered
        assert "Go north" in rendered
        assert "2." in rendered
        assert "Go south" in rendered
        assert "3." in rendered
        assert "Rest" in rendered
        assert "4." in rendered
        assert "Check inventory" in rendered

    def test_render_choices_empty(self):
        """Test rendering when no choices are available."""
        output = StringIO()
        console = Console(file=output, force_terminal=False, width=80)
        ui = UIRenderer(console)

        ui.render_choices([])

        rendered = output.getvalue()
        assert "No choices available" in rendered
        assert "end of the adventure" in rendered

    def test_render_choices_dict_format(self):
        """Test rendering choices in dict format."""
        output = StringIO()
        console = Console(file=output, force_terminal=False, width=80)
        ui = UIRenderer(console)

        # Test with dict-style choices (for backward compatibility)
        choices = [{"text": "Dict choice 1"}, {"text": "Dict choice 2"}]

        ui.render_choices(choices)

        rendered = output.getvalue()
        assert "Dict choice 1" in rendered
        assert "Dict choice 2" in rendered

    def test_render_inventory_with_items(self):
        """Test rendering inventory with items."""
        output = StringIO()
        console = Console(file=output, force_terminal=False, width=80)
        ui = UIRenderer(console)

        inventory = ["sword", "potion"]
        items_data = {
            "sword": {"name": "Magic Sword", "description": "A gleaming blade"},
            "potion": {"name": "Health Potion", "description": "Restores health"},
        }

        ui.render_inventory(inventory, items_data)

        rendered = output.getvalue()
        assert "Inventory" in rendered
        assert "Magic Sword" in rendered
        assert "A gleaming blade" in rendered
        assert "Health Potion" in rendered
        assert "Restores health" in rendered

    def test_render_inventory_empty(self):
        """Test rendering empty inventory."""
        output = StringIO()
        console = Console(file=output, force_terminal=False, width=80)
        ui = UIRenderer(console)

        ui.render_inventory([], {})

        rendered = output.getvalue()
        assert "Inventory" in rendered
        assert "empty" in rendered

    def test_render_inventory_missing_item_data(self):
        """Test rendering inventory with missing item data."""
        output = StringIO()
        console = Console(file=output, force_terminal=False, width=80)
        ui = UIRenderer(console)

        inventory = ["unknown_item"]
        items_data = {}

        ui.render_inventory(inventory, items_data)

        rendered = output.getvalue()
        assert "unknown_item" in rendered
        assert "No description" in rendered

    def test_render_action_history_with_actions(self):
        """Test rendering action history with actions."""
        output = StringIO()
        console = Console(file=output, force_terminal=False, width=80)
        ui = UIRenderer(console)

        action_log = [
            ActionLogEntry(1, "make_choice", "Made choice: Go north"),
            ActionLogEntry(2, "update_state", "Moved to new location"),
            ActionLogEntry(3, "take_item", "Took sword"),
        ]

        ui.render_action_history(action_log)

        rendered = output.getvalue()
        assert "Recent Actions" in rendered
        assert "Make Choice" in rendered  # Should be title-cased
        assert "Made choice: Go north" in rendered
        assert "Update State" in rendered
        assert "Take Item" in rendered

    def test_render_action_history_empty(self):
        """Test rendering empty action history."""
        output = StringIO()
        console = Console(file=output, force_terminal=False, width=80)
        ui = UIRenderer(console)

        ui.render_action_history([])

        rendered = output.getvalue()
        assert "No history available" in rendered

    def test_render_action_history_with_limit(self):
        """Test rendering action history with limit."""
        output = StringIO()
        console = Console(file=output, force_terminal=False, width=80)
        ui = UIRenderer(console)

        # Create more actions than limit
        action_log = [
            ActionLogEntry(i, f"action_{i}", f"Action {i}") for i in range(15)
        ]

        ui.render_action_history(action_log, limit=5)

        rendered = output.getvalue()
        # Should only show last 5 actions
        assert "Action 14" in rendered  # Last action
        assert "Action 10" in rendered  # 5th from last
        assert "Action 9" not in rendered  # 6th from last should not be shown
        assert "more actions" in rendered  # Should indicate there are more

    def test_render_status_basic(self):
        """Test rendering basic status."""
        output = StringIO()
        console = Console(file=output, force_terminal=False, width=80)
        ui = UIRenderer(console)

        current_event = Event("current", "Current event", location="Test Location")
        game_state = GameState(
            current_event=current_event, inventory=["item1", "item2"], act=1, scene=2
        )
        event_history = [Event("event1", "Event 1"), current_event]

        ui.render_status(game_state, event_history)

        rendered = output.getvalue()
        assert "Game Status" in rendered
        assert "Events Visited" in rendered
        assert "2" in rendered  # 2 events
        assert "Items in Inventory" in rendered
        assert "2" in rendered  # 2 items
        assert "Current Act" in rendered
        assert "2" in rendered  # Act 2 (1-indexed)
        assert "Current Scene" in rendered
        assert "3" in rendered  # Scene 3 (1-indexed)
        assert "Test Location" in rendered

    def test_render_status_with_story_data(self):
        """Test rendering status with story data for scene description."""
        output = StringIO()
        console = Console(file=output, force_terminal=False, width=80)
        ui = UIRenderer(console)

        game_state = GameState(act=0, scene=0)
        event_history = []
        story_data = StoryData(
            acts=[
                {
                    "name": "Act 1",
                    "description": "The beginning",
                    "scenes": [
                        {
                            "name": "Opening Scene",
                            "description": "A detailed scene description",
                            "location": "Starting location",
                        }
                    ],
                }
            ]
        )

        ui.render_status(game_state, event_history, story_data)

        rendered = output.getvalue()
        assert "Current Scene" in rendered
        assert "Opening Scene" in rendered
        assert "A detailed scene description" in rendered

    def test_render_choice_feedback(self):
        """Test rendering choice feedback."""
        output = StringIO()
        console = Console(file=output, force_terminal=False, width=80)
        ui = UIRenderer(console)

        ui.render_choice_feedback("Go north")

        rendered = output.getvalue()
        assert "> Go north" in rendered

    def test_render_error(self):
        """Test rendering error message."""
        output = StringIO()
        console = Console(file=output, force_terminal=False, width=80)
        ui = UIRenderer(console)

        ui.render_error("Test error message")

        rendered = output.getvalue()
        assert "✗ Test error message" in rendered

    def test_render_success(self):
        """Test rendering success message."""
        output = StringIO()
        console = Console(file=output, force_terminal=False, width=80)
        ui = UIRenderer(console)

        ui.render_success("Test success message")

        rendered = output.getvalue()
        assert "✓ Test success message" in rendered

    def test_render_info(self):
        """Test rendering info message."""
        output = StringIO()
        console = Console(file=output, force_terminal=False, width=80)
        ui = UIRenderer(console)

        ui.render_info("Test info message")

        rendered = output.getvalue()
        assert "Test info message" in rendered

    def test_render_goodbye(self):
        """Test rendering goodbye message."""
        output = StringIO()
        console = Console(file=output, force_terminal=False, width=80)
        ui = UIRenderer(console)

        ui.render_goodbye()

        rendered = output.getvalue()
        assert "Goodbye!" in rendered
        assert "Thanks for playing" in rendered

    def test_render_status_no_current_event(self):
        """Test rendering status without current event."""
        output = StringIO()
        console = Console(file=output, force_terminal=False, width=80)
        ui = UIRenderer(console)

        game_state = GameState()  # No current event
        event_history = []

        ui.render_status(game_state, event_history)

        rendered = output.getvalue()
        assert "Game Status" in rendered
        assert "Unknown" in rendered  # Current location should be Unknown

    def test_render_status_missing_scene_data(self):
        """Test rendering status with missing scene data."""
        output = StringIO()
        console = Console(file=output, force_terminal=False, width=80)
        ui = UIRenderer(console)

        game_state = GameState(act=0, scene=1)  # Scene 1 but only scene 0 exists
        event_history = []
        story_data = StoryData(
            acts=[
                {
                    "name": "Act 1",
                    "scenes": [{"name": "Scene 0", "description": "Only scene"}],
                }
            ]
        )

        ui.render_status(game_state, event_history, story_data)

        rendered = output.getvalue()
        # Should not crash, just not show scene description
        assert "Game Status" in rendered



"""Game engine core logic for Texventure."""

import os
from typing import List, Optional, Tuple

from .event_generators import LLMEventGenerator, TemplateEventGenerator
from .models import ActionLogEntry, Choice, Event, GameState, StoryData
from .save_manager import SaveLoadManager


class GameEngine:
    """Core game engine handling game logic and state management."""

    def __init__(
        self, use_llm: bool = True, model: str = "gpt-4.1-nano", max_tokens=800
    ):
        self.story_data: Optional[StoryData] = None
        self.game_state = GameState()
        self.event_history: List[Event] = []
        self.action_log: List[ActionLogEntry] = []
        self.event_counter = 0
        self.use_llm = use_llm
        self.max_tokens = max_tokens
        self.model = model
        self.save_manager = SaveLoadManager()
        self._setup_event_generator()

    def _setup_event_generator(self) -> None:
        """Setup the appropriate event generator based on configuration."""
        if self.use_llm and (api_key := os.getenv("OPENAI_API_KEY")):
            self.event_generator = LLMEventGenerator(api_key, self.model)
        else:
            self.event_generator = TemplateEventGenerator()

    def load_story(self, story_data: dict) -> None:
        """Load story data into the game engine."""
        self.story_data = StoryData(
            title=story_data.get("title", "Text Adventure"),
            setting=story_data.get("setting", "Unknown"),
            outline=story_data.get("outline", ""),
            acts=story_data.get("acts", []),
            npcs=story_data.get("npcs", []),
            items=story_data.get("items", {}),
        )

        if self._is_saved_game(story_data):
            self._load_saved_game(story_data)
        else:
            self._initialize_game_state()

    def _is_saved_game(self, story_data: dict) -> bool:
        """Check if the data contains a saved game state."""
        return "game_state" in story_data and "event_history" in story_data

    def _load_saved_game(self, save_data: dict) -> None:
        """Load a saved game state."""
        self.game_state, self.event_history, self.action_log, self.event_counter = (
            self.save_manager.load_saved_game(save_data)
        )

    def _initialize_game_state(self) -> None:
        """Initialize the game state."""
        self.game_state = GameState(act=0, scene=0)
        self.event_counter = 0

    def log_action(self, action_type: str, details: str) -> None:
        """Log an action taken by the player."""
        self.action_log.append(
            ActionLogEntry(
                timestamp=self.event_counter, action_type=action_type, details=details
            )
        )

    def generate_event(self, choice_text: str = "begin") -> Event:
        """Generate a new event based on the current game state and choice."""
        if not self.story_data:
            raise ValueError("No story data loaded")

        context = self._build_generation_context()
        event_id = f"event_{self.event_counter}"
        return self.event_generator.generate_event(context, choice_text, event_id)

    def _build_generation_context(self) -> dict:
        """Build context dictionary for event generation."""
        context = {
            "story_data": self.story_data,
            "game_state": self.game_state,
            "current_event": self.game_state.current_event,
        }

        # Add previous event if available
        if len(self.event_history) >= 2:
            context["previous_event"] = self.event_history[-2]

        return context

    def initial_event(self) -> Event:
        """Generate the initial event to start the game."""
        if not self.story_data or not self.story_data.acts:
            raise ValueError("No acts defined in story data")

        # Start with the first act and scene
        self.game_state.act = 0
        self.game_state.scene = 0
        event = self.generate_event("begin")
        self.game_state.current_event = event
        return event

    def update_state(self, choice_idx: int) -> Event:
        """Update the game state based on a choice and return the new event."""
        self.event_counter += 1
        current_event = self.game_state.current_event

        # Check if this is the final trigger before advancing
        is_final_trigger = (choice_idx == current_event.trigger_choice_idx) and (
            self._is_at_final_scene()
        )

        # Handle scene/act progression if choice is a trigger
        if (choice_idx) == current_event.trigger_choice_idx:
            self._advance_scene_and_act()

        # Check if game has ended after triggering the final trigger
        if is_final_trigger:
            # Create final event with no choices
            final_event = Event(
                id=f"event_{self.event_counter}_final",
                description="The adventure comes to an end. Thank you for playing!",
                choices=[],  # No choices - game over
                location=current_event.location if current_event else "Final location",
                act=self.game_state.act,
                scene=self.game_state.scene,
                trigger_choice_idx=-1,
            )
            self._update_game_state_with_event(final_event)
            self.log_action("game_end", "Game completed")
            return final_event

        # Generate new event based on choice
        choice_text = current_event.choices[choice_idx].text
        new_event = self.generate_event(choice_text)
        print(new_event)

        # Update game state
        self._update_game_state_with_event(new_event)
        self.log_action("make_choice", f"Made choice: {choice_text}")

        return new_event

    def _advance_scene_and_act(self) -> None:
        """Advance to the next scene and act if necessary."""
        if not self.story_data or not self.story_data.acts:
            return

        current_act = self.story_data.acts[self.game_state.act]
        scenes = current_act.get("scenes", [])

        if scenes:
            # Check if we're at the last scene of the last act
            is_last_act = self.game_state.act == len(self.story_data.acts) - 1
            is_last_scene = self.game_state.scene == len(scenes) - 1

            if is_last_act and is_last_scene:
                # Don't advance beyond the final scene
                return

            self.game_state.scene = (self.game_state.scene + 1) % len(scenes)
            if self.game_state.scene == 0:
                self.game_state.act = (self.game_state.act + 1) % len(
                    self.story_data.acts
                )

    def _is_at_final_scene(self) -> bool:
        """Check if we're currently at the final scene (before triggering the end)."""
        if not self.story_data or not self.story_data.acts:
            return False

        # Check if we're at the last act and last scene
        last_act_index = len(self.story_data.acts) - 1
        if self.game_state.act < last_act_index:
            return False

        current_act = self.story_data.acts[self.game_state.act]
        scenes = current_act.get("scenes", [])
        if not scenes:
            return False

        last_scene_index = len(scenes) - 1
        return self.game_state.scene >= last_scene_index

    def _update_game_state_with_event(self, event: Event) -> None:
        """Update game state with the new event."""
        self.game_state.current_event = event
        self.event_history.append(event)

        if event.id not in self.game_state.visited_events:
            self.game_state.visited_events.append(event.id)

    def make_choice(self, choice_index: int) -> Tuple[bool, str, Optional[Event]]:
        """Make a choice and return success status, message, and new event."""
        if not self.game_state.current_event:
            return False, "No current event", None

        choices = self.game_state.current_event.choices

        # Check if game has ended (no choices available)
        if len(choices) == 0:
            return False, "The game has ended. No more choices available.", None

        if not self._is_valid_choice_index(choice_index, len(choices)):
            return False, f"Invalid choice. Choose between 1 and {len(choices)}", None

        choice = choices[choice_index]
        new_event = self.update_state(choice_index)
        return True, f"You chose: {choice.text}", new_event

    def _is_valid_choice_index(self, choice_index: int, num_choices: int) -> bool:
        """Check if the choice index is valid."""
        return 0 <= choice_index < num_choices

    def is_game_ended(self) -> bool:
        """Public method to check if the game has ended."""
        # Game has ended if current event has no choices
        current_event = self.game_state.current_event
        return current_event is not None and len(current_event.choices) == 0

    def add_to_inventory(self, item_id: str) -> Tuple[bool, str]:
        """Add an item to inventory."""
        # Validate item availability
        if validation_error := self._validate_item_availability(item_id):
            return False, validation_error

        # Add to inventory and remove from event
        self.game_state.inventory.append(item_id)
        self.game_state.current_event.items.remove(item_id)

        item_name = self.story_data.items[item_id].get("name", item_id)
        self.log_action("take_item", f"Took item: {item_name} ({item_id})")
        return True, f"You take the {item_name}"

    def _validate_item_availability(self, item_id: str) -> Optional[str]:
        """Validate if an item can be taken. Returns error message or None."""
        if not self.story_data or not self.story_data.items:
            return "No items available"

        if item_id not in self.story_data.items:
            return f"Item '{item_id}' not found"

        if item_id in self.game_state.inventory:
            return f"You already have {item_id}"

        current_event = self.game_state.current_event
        if not current_event or item_id not in current_event.items:
            return f"'{item_id}' is not available here"

        return None

    def save_game(self, filename: str) -> Tuple[bool, str]:
        """Save the current game state."""
        success, message = self.save_manager.save_game(
            filename,
            self.story_data,
            self.game_state,
            self.event_history,
            self.action_log,
            self.event_counter,
        )

        if success:
            self.log_action("save_game", f"Game saved to: {filename}")

        return success, message



"""Tests for event generators."""

import json
from unittest.mock import Mock, patch

import pytest

from src.texventure.event_generators import (EventGenerator, LLMEventGenerator,
                                             TemplateEventGenerator)
from src.texventure.models import Choice, Event, GameState, StoryData


class TestTemplateEventGenerator:
    """Test cases for TemplateEventGenerator."""

    def test_generate_begin_event_with_story(self, template_event_generator):
        """Test generating the begin event with story data."""
        story_data = StoryData(
            title="Test Adventure",
            acts=[
                {
                    "name": "Act 1",
                    "description": "The beginning",
                    "scenes": [
                        {
                            "name": "Scene 1",
                            "description": "Starting scene description",
                            "location": "Starting location",
                        }
                    ],
                }
            ],
        )

        context = {"story_data": story_data, "game_state": GameState()}

        event = template_event_generator.generate_event(context, "begin", "event_0")

        assert event.id == "event_0"
        assert (
            "Starting scene description" in event.description
            or "The beginning" in event.description
        )
        assert event.location in ["Starting location", "Unknown location"]
        assert len(event.choices) == 4
        assert event.act == 0
        assert event.scene == 0

    def test_generate_begin_event_without_story(self, template_event_generator):
        """Test generating the begin event without story data."""
        context = {"story_data": None, "game_state": GameState()}

        event = template_event_generator.generate_event(context, "begin", "event_0")

        assert event.id == "event_0"
        assert "You chose: begin" in event.description
        assert event.location == "Current location"
        assert len(event.choices) == 4

    def test_generate_continuation_event(self, template_event_generator):
        """Test generating a continuation event."""
        context = {"story_data": StoryData(), "game_state": GameState(act=1, scene=2)}

        event = template_event_generator.generate_event(context, "go north", "event_1")

        assert event.id == "event_1"
        assert "You chose: go north" in event.description
        assert event.location == "Current location"
        assert len(event.choices) == 4
        assert event.act == 1
        assert event.scene == 2

    def test_generate_event_choice_types(self, template_event_generator):
        """Test that generated events have correct choice types."""
        context = {"story_data": StoryData(), "game_state": GameState()}

        event = template_event_generator.generate_event(context, "test", "event_0")

        for choice in event.choices:
            assert isinstance(choice, Choice)
            assert isinstance(choice.text, str)
            assert len(choice.text) > 0


class TestLLMEventGenerator:
    """Test cases for LLMEventGenerator."""

    def test_llm_generator_initialization(self):
        """Test LLM generator initialization."""
        api_key = "test_key"
        model = "gpt-4"
        generator = LLMEventGenerator(api_key, model)

        assert generator.api_key == api_key
        assert generator.model == model

    def test_llm_generator_default_model(self):
        """Test LLM generator with default model."""
        generator = LLMEventGenerator("test_key")
        assert generator.model == "gpt-3.5-turbo"

    @patch("src.texventure.event_generators.llm_call")
    def test_generate_begin_event_success(self, mock_llm_call, mock_llm_response):
        """Test successful begin event generation with LLM."""
        mock_llm_call.return_value = mock_llm_response

        generator = LLMEventGenerator("test_key")
        context = {
            "story_data": StoryData(title="Test", setting="Test world"),
            "game_state": GameState(),
        }

        event = generator.generate_event(context, "begin", "event_0")

        assert event.id == "event_0"
        assert "mysterious forest path" in event.description.lower()
        assert event.location == "Misty Forest Path"
        assert len(event.choices) == 4
        mock_llm_call.assert_called_once()

        # Check that force_json=True was passed
        call_args = mock_llm_call.call_args
        assert call_args[1]["force_json"] is True

    @patch("src.texventure.event_generators.llm_call")
    def test_generate_continuation_event_success(
        self, mock_llm_call, mock_llm_response
    ):
        """Test successful continuation event generation with LLM."""
        mock_llm_call.return_value = mock_llm_response

        generator = LLMEventGenerator("test_key")
        context = {
            "story_data": StoryData(title="Test", setting="Test world"),
            "game_state": GameState(),
            "current_event": Event("prev_event", "Previous event"),
            "previous_event": Event("prev_prev_event", "Previous previous event"),
        }

        event = generator.generate_event(context, "go north", "event_1")

        assert event.id == "event_1"
        assert len(event.choices) == 4
        mock_llm_call.assert_called_once()

    @patch("src.texventure.event_generators.llm_call")
    def test_generate_event_llm_failure(self, mock_llm_call):
        """Test event generation when LLM call fails."""
        mock_llm_call.side_effect = Exception("API Error")

        generator = LLMEventGenerator("test_key")
        context = {"story_data": StoryData(), "game_state": GameState()}

        event = generator.generate_event(context, "test", "event_0")

        # Should fall back to template event
        assert event.id == "event_0"
        assert "adventure continues" in event.description.lower()
        assert len(event.choices) == 4

    @patch("src.texventure.event_generators.llm_call")
    def test_generate_event_invalid_json(self, mock_llm_call):
        """Test event generation with invalid JSON response."""
        mock_llm_call.return_value = "Invalid JSON response"

        generator = LLMEventGenerator("test_key")
        context = {"story_data": StoryData(), "game_state": GameState()}

        event = generator.generate_event(context, "test", "event_0")

        # Should fall back to template event
        assert event.id == "event_0"
        assert len(event.choices) == 4

    def test_build_context_string_comprehensive(self):
        """Test building comprehensive context string."""
        generator = LLMEventGenerator("test_key")

        story_data = StoryData(
            title="Test Adventure",
            setting="Fantasy realm",
            outline="Epic quest",
            acts=[
                {
                    "name": "Act 1",
                    "description": "The beginning",
                    "scenes": [
                        {
                            "name": "Scene 1",
                            "description": "Starting area",
                            "location": "Village",
                        }
                    ],
                }
            ],
            npcs=[{"name": "Wizard", "description": "Wise old wizard"}],
        )

        game_state = GameState(inventory=["sword", "potion"], act=0, scene=0)

        current_event = Event(
            "current_id",
            "Current event description",
            location="Current location",
            items=["treasure"],
        )

        previous_event = Event(
            "prev_id", "Previous event description", location="Previous location"
        )

        context = {
            "story_data": story_data,
            "game_state": game_state,
            "current_event": current_event,
            "previous_event": previous_event,
        }

        context_str = generator._build_context_string(context)

        assert "Test Adventure" in context_str
        assert "Fantasy realm" in context_str
        assert "Epic quest" in context_str
        assert "Act 1" in context_str
        assert "The beginning" in context_str
        assert "Scene 1" in context_str
        assert "Starting area" in context_str
        assert "Village" in context_str
        assert "Wizard" in context_str
        assert "Wise old wizard" in context_str
        assert "Current event description" in context_str
        assert "Current location" in context_str
        assert "treasure" in context_str
        assert "Previous event description" in context_str
        assert "Previous location" in context_str
        assert "sword, potion" in context_str

    def test_build_context_string_minimal(self):
        """Test building context string with minimal data."""
        generator = LLMEventGenerator("test_key")

        context = {}
        context_str = generator._build_context_string(context)

        # Should not crash and return empty or minimal string
        assert isinstance(context_str, str)

    def test_parse_llm_response_valid(self):
        """Test parsing valid LLM response."""
        generator = LLMEventGenerator("test_key")

        response = json.dumps(
            {
                "description": "Test description",
                "location": "Test location",
                "choices": [
                    {"text": "Choice 1"},
                    {"text": "Choice 2"},
                    {"text": "Choice 3"},
                    {"text": "Choice 4"},
                ],
            }
        )

        context = {"game_state": GameState(act=1, scene=2)}
        event = generator._parse_llm_response(response, "test_id", context)

        assert event.id == "test_id"
        assert event.description == "Test description"
        assert event.location == "Test location"
        assert len(event.choices) == 4
        assert event.choices[0].text == "Choice 1"
        assert event.act == 1
        assert event.scene == 2

    def test_parse_llm_response_missing_description(self):
        """Test parsing LLM response with missing description."""
        generator = LLMEventGenerator("test_key")

        response = json.dumps(
            {"location": "Test location", "choices": [{"text": "Choice 1"}]}
        )

        context = {"game_state": GameState()}
        event = generator._parse_llm_response(response, "test_id", context)

        # Should fall back to default event
        assert event.id == "test_id"
        assert "adventure continues" in event.description.lower()
        assert len(event.choices) == 4

    def test_parse_llm_response_empty_choices(self):
        """Test parsing LLM response with empty choices."""
        generator = LLMEventGenerator("test_key")

        response = json.dumps(
            {
                "description": "Test description",
                "location": "Test location",
                "choices": [],
            }
        )

        context = {"game_state": GameState()}
        event = generator._parse_llm_response(response, "test_id", context)

        assert event.description == "Test description"
        assert event.location == "Test location"
        # Should have default choices
        assert len(event.choices) == 4

    def test_create_fallback_event(self):
        """Test creating fallback event."""
        generator = LLMEventGenerator("test_key")

        context = {"game_state": GameState(act=2, scene=1)}
        event = generator._create_fallback_event("fallback_id", "test_choice", context)

        assert event.id == "fallback_id"
        assert "adventure continues" in event.description.lower()
        assert event.location == "Unknown location"
        assert len(event.choices) == 4
        assert event.act == 2
        assert event.scene == 1

    def test_prompt_generation_begin(self):
        """Test that begin prompts are generated correctly."""
        generator = LLMEventGenerator("test_key")

        context_str = "Test context"
        prompt = generator._create_opening_prompt(context_str)

        assert "opening event" in prompt.lower()
        assert "interactive text adventure" in prompt.lower()
        assert context_str in prompt
        assert "valid JSON" in prompt

    def test_prompt_generation_continuation(self):
        """Test that continuation prompts are generated correctly."""
        generator = LLMEventGenerator("test_key")

        context_str = "Test context"
        choice_text = "go north"
        prompt = generator._create_continuation_prompt(context_str, choice_text)

        assert "next event" in prompt.lower()
        assert "interactive text adventure" in prompt.lower()
        assert context_str in prompt
        assert choice_text in prompt
        assert "valid JSON" in prompt



"""Story generator module for Texventure."""

import argparse
import json
import os
import sys
from typing import List

import requests
import toml
from pydantic import BaseModel, Field


class Scene(BaseModel):
    name: str = ""
    description: str = ""
    location: str = ""
    trigger: str = ""


class Act(BaseModel):
    name: str = ""
    description: str = ""
    scenes: List[Scene] = Field(default_factory=list)


class NPC(BaseModel):
    name: str = ""
    role: str = ""
    description: str = ""
    personality: str = ""


class Story(BaseModel):
    title: str = ""
    setting: str = ""
    outline: str = ""
    style_genre: str = ""
    n_acts: int = 3
    n_scenes: int = 2
    n_npcs: int = 3
    npcs: List[NPC] = Field(default_factory=list)
    acts: List[Act] = Field(default_factory=list)


def llm_call(prompt, api_key, model="gpt-4.1-nano", force_json=False, max_tokens=800):
    """Call the LLM API to generate the story."""
    api_url = "https://api.openai.com/v1/chat/completions"
    headers = {"Content-Type": "application/json", "Authorization": f"Bearer {api_key}"}

    # Adjust system message based on whether JSON is requested
    if force_json:
        system_content = "You are a creative writing assistant that generates detailed interactive fiction stories. Always respond with valid JSON when requested, using the exact field names provided in the prompt."
    else:
        system_content = "You are a creative writing assistant that generates detailed interactive fiction stories."
    payload = {
        "model": model,
        "messages": [
            {
                "role": "system",
                "content": system_content,
            },
            {"role": "user", "content": prompt},
        ],
        "max_tokens": max_tokens,
        "temperature": 0.8,
    }

    # Add JSON response format if requested
    if force_json:
        payload["response_format"] = {"type": "json_object"}

    try:
        response = requests.post(api_url, headers=headers, json=payload)
        response.raise_for_status()
        result = response.json()
        content = result["choices"][0]["message"]["content"]
        return content.strip() if content else "No response from LLM"

    except requests.RequestException as e:
        print(f"Error calling LLM API: {e}")
        return None


def form_bulk_prompt(empty_fields, context):
    """Form a prompt to fill multiple empty fields at once."""
    fields_list = "\n".join([f"- {field}" for field in empty_fields])
    return f"""Generate content for the following empty fields in an interactive fiction story. Return your response as a JSON object where each key is the field path and the value is the generated content.

        Each npc is a character in the story, they have a name, role, description, and personality.
        Each act is a major section of the story with a name and description.
        Each scene is a part of an act with a name, description, location, and trigger.
        Descriptions are detailed and should fit the context of the story and describe the narrative.
        Triggers are actions or choices the user makes that lead to the next scene in the story.
        
        All content should match the setting, outline, and style_genre specified in the story context. The style_genre defines the narrative style and genre conventions that should be followed throughout the story.

        Empty fields to fill:
        {fields_list}

        Current story data for context:
        {context}

        Return a JSON object with the field paths as keys and the generated content as values. For example:
        {{"title": "The Mysterious Castle", "setting": "A dark medieval fantasy world", "outline": "A hero must explore a haunted castle to rescue a princess", "style_genre": "Gothic horror adventure"}}"""


def collect_empty_fields(data, path=""):
    """Collect all empty string fields in the data structure."""
    empty_fields = []

    if isinstance(data, dict):
        for key, value in data.items():
            current_path = f"{path}.{key}" if path else key
            if isinstance(value, (dict, list)):
                empty_fields.extend(collect_empty_fields(value, current_path))
            elif value == "":
                empty_fields.append(current_path)
    elif isinstance(data, list):
        for i, item in enumerate(data):
            current_path = f"{path}[{i}]"
            empty_fields.extend(collect_empty_fields(item, current_path))

    return empty_fields


def set_nested_value(data, path, value):
    """Set a value in nested data structure using dot notation path."""
    parts = path.split(".")
    current = data

    for part in parts[:-1]:
        if "[" in part and "]" in part:
            # Handle array notation like "acts[0]"
            key, index_str = part.split("[")
            index = int(index_str.rstrip("]"))
            current = current[key][index]
        else:
            current = current[part]

    final_part = parts[-1]
    if "[" in final_part and "]" in final_part:
        # Handle array notation for final part
        key, index_str = final_part.split("[")
        index = int(index_str.rstrip("]"))
        current[key][index] = value
    else:
        current[final_part] = value


def generate_story(story_data, api_key, model="gpt-4.1-nano", max_tokens=800):
    """Generate story content using JSON responses to fill multiple fields at once."""

    if not api_key:
        print("No API key provided, skipping LLM generation")
        return

    # Collect all empty fields
    empty_fields = collect_empty_fields(story_data)

    if not empty_fields:
        print("No empty fields to fill")
        return

    print(f"Found {len(empty_fields)} empty fields to fill")

    # Process fields in batches to avoid token limits
    batch_size = 10  # Adjust based on token limits and field complexity

    for i in range(0, len(empty_fields), batch_size):
        batch_fields = empty_fields[i : i + batch_size]
        print(f"Filling batch {i//batch_size + 1}: {len(batch_fields)} fields")

        # Create prompt for this batch
        context = json.dumps(story_data, indent=2)
        prompt = form_bulk_prompt(batch_fields, context)

        # Get JSON response from LLM
        llm_response = llm_call(
            prompt, api_key, model, force_json=True, max_tokens=max_tokens
        )

        if llm_response:
            try:
                # Parse JSON response
                generated_content = json.loads(llm_response)

                # Apply generated content to story data
                for field_path, content in generated_content.items():
                    if field_path in batch_fields and content.strip():
                        set_nested_value(story_data, field_path, content.strip())
                        print(f"  Filled: {field_path}")

            except json.JSONDecodeError as e:
                print(f"Failed to parse JSON response: {e}")
                print(f"Raw response: {llm_response}")
                # Fallback to single-field approach for this batch
                _fill_fields_individually(
                    story_data, batch_fields, api_key, model, max_tokens
                )
        else:
            print(f"No response from LLM for batch {i//batch_size + 1}")


def _fill_fields_individually(story_data, fields, api_key, model, max_tokens=800):
    """Fallback method to fill fields one by one if JSON parsing fails."""
    print("Falling back to individual field filling...")

    for field_path in fields:
        prompt = f"Generate content for the '{field_path}' field of an interactive fiction story. Ensure the content matches the setting, outline, and style_genre specified in the story context. The current story data is: {json.dumps(story_data, indent=2)}"
        llm_response = llm_call(prompt, api_key, model, max_tokens=max_tokens)

        if llm_response and llm_response.strip():
            set_nested_value(story_data, field_path, llm_response.strip())
            print(f"  Filled: {field_path}")


def format_story(template_data):
    story = Story.model_validate(template_data)
    story_data = story.model_dump()
    # Ensure we have the required number of acts
    while len(story_data["acts"]) < story_data["n_acts"]:
        act_number = len(story_data["acts"]) + 1
        new_act = Act(name="", description="", scenes=[])
        story_data["acts"].append(new_act.model_dump())

    # Ensure each act has the required number of scenes
    for act in story_data["acts"]:
        while len(act["scenes"]) < story_data["n_scenes"]:
            scene_number = len(act["scenes"]) + 1
            new_scene = Scene(name="", description="", location="", trigger="")
            act["scenes"].append(new_scene.model_dump())

    # Ensure we have the required number of NPCs
    while len(story_data["npcs"]) < story_data["n_npcs"]:
        npc_number = len(story_data["npcs"]) + 1
        new_npc = NPC(name="", role="", description="", personality="")
        story_data["npcs"].append(new_npc.model_dump())
    return story_data


def main():
    """Main entry point for the texventure-generate command."""
    parser = argparse.ArgumentParser(
        description="Generate a story file from a TOML template"
    )
    parser.add_argument("toml_file", help="Path to the TOML template file")
    parser.add_argument(
        "-o",
        "--output",
        help="Output JSON file path (default: story.json)",
        default="story.json",
    )
    parser.add_argument(
        "--api-key", help="OpenAI API key (or set OPENAI_API_KEY environment variable)"
    )
    parser.add_argument("--model", help="LLM model to use", default="gpt-4.1-nano")
    parser.add_argument(
        "--max-tokens",
        type=int,
        help="Maximum number of tokens for LLM responses (default: 800)",
        default=800,
    )
    parser.add_argument(
        "--no-llm", action="store_true", help="Generate basic story without LLM"
    )
    args = parser.parse_args()

    # Read TOML template file
    try:
        with open(args.toml_file, "r") as f:
            template_data = toml.load(f)

    except FileNotFoundError:
        print(f"Error: Template file {args.toml_file} not found")
        return 1

    print(f"Generating story from template: {args.toml_file}")

    story_data = format_story(template_data)

    if not args.no_llm:
        # Get API key
        api_key = args.api_key or os.getenv("OPENAI_API_KEY")

        if api_key:
            print("Calling LLM to generate detailed story...")
            generate_story(story_data, api_key, args.model, args.max_tokens)

        else:
            print(
                "Warning: No API key provided. Use --api-key or set OPENAI_API_KEY environment variable."
            )
    else:
        print("Generating basic outline without LLM...")

    # Write story to output file
    try:
        with open(args.output, "w") as f:
            json.dump(story_data, f)
        print(f"Story generated successfully: {args.output}")
        print(f"Run the game with: texventure {args.output}")
        return 0
    except IOError as e:
        print(f"Error writing output file {args.output}: {e}")
        return 1


if __name__ == "__main__":
    sys.exit(main())



"""Integration tests for Texventure."""

import json
import os
import tempfile
from unittest.mock import patch

import pytest

from src.texventure.engine import GameEngine
from src.texventure.main import TexventureShell
from src.texventure.models import Choice, Event


class TestIntegration:
    """Integration tests for the complete Texventure system."""

    def test_complete_game_flow(self, sample_story_data):
        """Test a complete game flow from start to finish."""
        # Create a temporary story file
        with tempfile.NamedTemporaryFile(mode="w", delete=False, suffix=".json") as f:
            json.dump(sample_story_data, f)
            story_file = f.name

        try:
            # Initialize shell
            shell = TexventureShell(use_llm=False)  # Use template to avoid LLM calls

            # Load story
            with (
                patch.object(shell.ui, "render_title_screen"),
                patch.object(shell.ui, "render_event"),
            ):
                shell.load_story(story_file)

            # Verify story loaded
            assert shell.engine.story_data.title == "Test Adventure"
            assert shell.engine.game_state.current_event is not None
            assert len(shell.engine.event_history) == 1

            # Make a choice
            with (
                patch.object(shell.ui, "render_choice_feedback"),
                patch.object(shell.ui, "render_event"),
            ):
                shell.do_choice("1")

            # Verify choice was made
            assert len(shell.engine.event_history) == 2
            assert len(shell.engine.action_log) >= 2

            # Check inventory initially empty
            with patch.object(shell.ui, "render_inventory"):
                shell.do_inventory("")

            # Try to take an item (if available)
            current_event = shell.engine.game_state.current_event
            if current_event and current_event.items:
                item_to_take = current_event.items[0]
                with patch.object(shell.ui, "render_success"):
                    shell.do_take(item_to_take)

                # Verify item was taken
                assert item_to_take in shell.engine.game_state.inventory
                assert item_to_take not in current_event.items

            # Save game
            save_file = f"test-save-{os.getpid()}.json"
            try:
                with (
                    patch.object(shell.ui, "render_success"),
                    patch.object(shell.ui, "render_info"),
                ):
                    shell.do_save(save_file)

                # Verify save file was created
                assert os.path.exists(save_file)

                # Load the saved game in a new shell
                shell2 = TexventureShell(use_llm=False)
                with (
                    patch.object(shell2.ui, "render_title_screen"),
                    patch.object(shell2.ui, "render_event"),
                    patch.object(shell2.console, "print"),
                ):
                    shell2.load_story(save_file)

                # Verify save was loaded correctly
                assert shell2.engine.event_counter == shell.engine.event_counter
                assert len(shell2.engine.event_history) == len(
                    shell.engine.event_history
                )
                assert (
                    shell2.engine.game_state.inventory
                    == shell.engine.game_state.inventory
                )

            finally:
                if os.path.exists(save_file):
                    os.unlink(save_file)

            # Check status
            with (
                patch.object(shell.ui, "render_status"),
                patch.object(shell.console, "print"),
            ):
                shell.do_status("")

            # Check history
            with patch.object(shell.ui, "render_action_history"):
                shell.do_history("")

            # Look around
            with patch.object(shell.ui, "render_event"):
                shell.do_look("")

        finally:
            if os.path.exists(story_file):
                os.unlink(story_file)

    def test_error_handling_integration(self):
        """Test error handling in integrated scenarios."""
        shell = TexventureShell(use_llm=False)

        # Try commands without loading story
        with patch.object(shell.ui, "render_error"):
            shell.do_take("sword")  # Should handle gracefully

        with patch.object(shell.ui, "render_error"):
            shell.do_choice("1")  # Should handle gracefully

        # Load invalid story file
        with patch.object(shell.ui, "render_error"), patch("sys.exit"):
            shell.load_story("nonexistent.json")

    def test_llm_fallback_integration(self, sample_story_data):
        """Test LLM fallback behavior in integration."""
        # Create story file
        with tempfile.NamedTemporaryFile(mode="w", delete=False, suffix=".json") as f:
            json.dump(sample_story_data, f)
            story_file = f.name

        try:
            # Test with LLM enabled but no API key
            with patch.dict("os.environ", {}, clear=True):
                shell = TexventureShell(use_llm=True)

                with (
                    patch.object(shell.ui, "render_title_screen"),
                    patch.object(shell.ui, "render_event"),
                ):
                    shell.load_story(story_file)

                # Should fall back to template generator
                from src.texventure.event_generators import \
                    TemplateEventGenerator

                assert isinstance(shell.engine.event_generator, TemplateEventGenerator)

                # Make choices should still work
                with (
                    patch.object(shell.ui, "render_choice_feedback"),
                    patch.object(shell.ui, "render_event"),
                ):
                    shell.do_choice("1")

                assert len(shell.engine.event_history) == 2

        finally:
            if os.path.exists(story_file):
                os.unlink(story_file)

    def test_save_load_roundtrip_integration(self, sample_story_data):
        """Test complete save/load roundtrip with complex state."""
        # Create story file
        with tempfile.NamedTemporaryFile(mode="w", delete=False, suffix=".json") as f:
            json.dump(sample_story_data, f)
            story_file = f.name

        try:
            # Create initial game state
            shell1 = TexventureShell(use_llm=False)

            with (
                patch.object(shell1.ui, "render_title_screen"),
                patch.object(shell1.ui, "render_event"),
            ):
                shell1.load_story(story_file)

            # Build up some game state
            for i in range(3):
                with (
                    patch.object(shell1.ui, "render_choice_feedback"),
                    patch.object(shell1.ui, "render_event"),
                ):
                    shell1.do_choice("1")

            # Take items if available
            for event in shell1.engine.event_history:
                if event.items:
                    for item in event.items[
                        :
                    ]:  # Copy to avoid modification during iteration
                        shell1.engine.game_state.current_event = event
                        with patch.object(shell1.ui, "render_success"):
                            shell1.do_take(item)

            # Save the complex state
            save_file = f"complex-save-{os.getpid()}.json"
            try:
                with (
                    patch.object(shell1.ui, "render_success"),
                    patch.object(shell1.ui, "render_info"),
                ):
                    shell1.do_save(save_file)

                # Load in new shell
                shell2 = TexventureShell(use_llm=False)
                with (
                    patch.object(shell2.ui, "render_title_screen"),
                    patch.object(shell2.ui, "render_event"),
                    patch.object(shell2.console, "print"),
                ):
                    shell2.load_story(save_file)

                # Verify all state was preserved
                assert shell2.engine.event_counter == shell1.engine.event_counter
                assert len(shell2.engine.event_history) == len(
                    shell1.engine.event_history
                )
                assert len(shell2.engine.action_log) == len(shell1.engine.action_log)
                assert (
                    shell2.engine.game_state.inventory
                    == shell1.engine.game_state.inventory
                )
                assert (
                    shell2.engine.game_state.visited_events
                    == shell1.engine.game_state.visited_events
                )

                # Continue playing from loaded state
                with (
                    patch.object(shell2.ui, "render_choice_feedback"),
                    patch.object(shell2.ui, "render_event"),
                ):
                    shell2.do_choice("2")

                # Verify game continues normally
                assert (
                    len(shell2.engine.event_history)
                    == len(shell1.engine.event_history) + 1
                )

            finally:
                if os.path.exists(save_file):
                    os.unlink(save_file)

        finally:
            if os.path.exists(story_file):
                os.unlink(story_file)

    def test_engine_event_generator_integration(self, sample_story_data):
        """Test integration between engine and event generators."""
        engine = GameEngine(use_llm=False)
        engine.load_story(sample_story_data)

        # Test event generation
        event = engine.generate_event("begin")
        assert isinstance(event, Event)
        assert event.id == "event_0"
        assert len(event.choices) == 4

        # Test state update
        new_event = engine.update_state("test choice")
        assert new_event != event
        assert new_event.id == "event_1"
        assert engine.game_state.current_event == new_event
        assert new_event in engine.event_history

        # Test choice making
        success, message, next_event = engine.make_choice(0)
        assert success is True
        assert next_event is not None
        assert len(engine.event_history) == 3  # begin + test choice + choice 0

    def test_ui_engine_integration(self, sample_story_data):
        """Test integration between UI and engine components."""
        shell = TexventureShell(use_llm=False)

        # Create story file
        with tempfile.NamedTemporaryFile(mode="w", delete=False, suffix=".json") as f:
            json.dump(sample_story_data, f)
            story_file = f.name

        try:
            # Load story and capture UI calls
            ui_calls = []

            def track_ui_call(method_name):
                def wrapper(*args, **kwargs):
                    ui_calls.append(method_name)
                    return None

                return wrapper

            # Patch UI methods to track calls
            with (
                patch.object(shell.ui, "render_title_screen", track_ui_call("title")),
                patch.object(shell.ui, "render_event", track_ui_call("event")),
                patch.object(
                    shell.ui, "render_choice_feedback", track_ui_call("choice_feedback")
                ),
                patch.object(shell.ui, "render_success", track_ui_call("success")),
                patch.object(shell.ui, "render_error", track_ui_call("error")),
                patch.object(shell.ui, "render_inventory", track_ui_call("inventory")),
                patch.object(shell.ui, "render_status", track_ui_call("status")),
                patch.object(
                    shell.ui, "render_action_history", track_ui_call("history")
                ),
                patch.object(shell.console, "print"),
            ):

                # Load story
                shell.load_story(story_file)
                assert "title" in ui_calls
                assert "event" in ui_calls

                # Make choice
                ui_calls.clear()
                shell.do_choice("1")
                assert "choice_feedback" in ui_calls
                assert "event" in ui_calls

                # Show inventory
                ui_calls.clear()
                shell.do_inventory("")
                assert "inventory" in ui_calls

                # Show status
                ui_calls.clear()
                shell.do_status("")
                assert "status" in ui_calls

                # Show history
                ui_calls.clear()
                shell.do_history("")
                assert "history" in ui_calls

                # Invalid choice
                ui_calls.clear()
                shell.do_choice("invalid")
                assert "error" in ui_calls

        finally:
            if os.path.exists(story_file):
                os.unlink(story_file)

    def test_cross_component_data_flow(self, sample_story_data):
        """Test data flow across all components."""
        engine = GameEngine(use_llm=False)
        engine.load_story(sample_story_data)

        # Trace data flow from story -> engine -> events -> choices -> state

        # 1. Story data flows to engine
        assert engine.story_data.title == sample_story_data["title"]
        assert len(engine.story_data.acts) == len(sample_story_data["acts"])

        # 2. Engine generates event using story context
        event = engine.generate_event("begin")
        assert event.description is not None
        assert len(event.choices) > 0

        # 3. Event updates engine state
        old_counter = engine.event_counter
        new_event = engine.update_state("test")
        assert engine.event_counter == old_counter + 1
        assert engine.game_state.current_event == new_event

        # 4. Choice affects state and generates new event
        choice_count = len(engine.event_history)
        action_count = len(engine.action_log)

        success, message, next_event = engine.make_choice(0)
        assert success is True
        assert len(engine.event_history) == choice_count + 1
        assert len(engine.action_log) > action_count

        # 5. Save captures all state
        save_data = {
            "title": engine.story_data.title,
            "setting": engine.story_data.setting,
            "acts": engine.story_data.acts,
            "npcs": engine.story_data.npcs,
            "items": engine.story_data.items,
            "is_saved_game": True,
            "event_counter": engine.event_counter,
            "game_state": {
                "inventory": engine.game_state.inventory,
                "act": engine.game_state.act,
                "scene": engine.game_state.scene,
                "visited_events": engine.game_state.visited_events,
                "flags": engine.game_state.flags,
            },
            "event_history": [
                {
                    "id": e.id,
                    "description": e.description,
                    "choices": [{"text": c.text} for c in e.choices],
                    "items": e.items,
                    "location": e.location,
                    "act": e.act,
                    "scene": e.scene,
                }
                for e in engine.event_history
            ],
            "action_log": [
                {
                    "timestamp": a.timestamp,
                    "action_type": a.action_type,
                    "details": a.details,
                }
                for a in engine.action_log
            ],
        }

        # 6. Load restores all state
        new_engine = GameEngine(use_llm=False)
        new_engine.load_story(save_data)

        assert new_engine.event_counter == engine.event_counter
        assert len(new_engine.event_history) == len(engine.event_history)
        assert len(new_engine.action_log) == len(engine.action_log)
        assert new_engine.game_state.inventory == engine.game_state.inventory



"""Tests for save manager."""

import json
import os
import tempfile
from unittest.mock import mock_open, patch

import pytest

from src.texventure.models import (ActionLogEntry, Choice, Event, GameState,
                                   StoryData)
from src.texventure.save_manager import SaveLoadManager


class TestSaveLoadManager:
    """Test cases for SaveLoadManager."""

    def test_save_game_success(self, save_manager, sample_story_data):
        """Test successful game saving."""
        story_data = StoryData(
            title="Test Game",
            setting="Test World",
            outline="Test outline",
            acts=sample_story_data["acts"],
            npcs=sample_story_data["npcs"],
            items=sample_story_data["items"],
        )

        current_event = Event(
            id="current_event",
            description="Current event description",
            choices=[Choice("Choice 1"), Choice("Choice 2")],
            items=["sword"],
            location="Current location",
            act=0,
            scene=1,
        )

        game_state = GameState(
            current_event=current_event,
            inventory=["potion"],
            act=0,
            scene=1,
            visited_events=["event_0", "current_event"],
            flags={"started": True},
        )

        event_history = [
            Event("event_0", "First event", [Choice("Start")]),
            current_event,
        ]

        action_log = [
            ActionLogEntry(0, "start_game", "Game started"),
            ActionLogEntry(1, "make_choice", "Made choice: Start"),
        ]

        event_counter = 2

        with tempfile.NamedTemporaryFile(mode="w", delete=False, suffix=".json") as f:
            filename = f.name

        try:
            success, message = save_manager.save_game(
                filename,
                story_data,
                game_state,
                event_history,
                action_log,
                event_counter,
            )

            assert success is True
            assert "Game saved to" in message
            assert filename in message

            # Verify file was created and contains expected data
            assert os.path.exists(filename)

            with open(filename, "r") as f:
                saved_data = json.load(f)

            # Check story data
            assert saved_data["title"] == "Test Game"
            assert saved_data["setting"] == "Test World"
            assert saved_data["outline"] == "Test outline"
            assert saved_data["acts"] == sample_story_data["acts"]
            assert saved_data["npcs"] == sample_story_data["npcs"]
            assert saved_data["items"] == sample_story_data["items"]

            # Check save metadata
            assert saved_data["is_saved_game"] is True
            assert saved_data["saved_at"] == event_counter
            assert saved_data["event_counter"] == event_counter

            # Check current event
            current_event_data = saved_data["current_event"]
            assert current_event_data["id"] == "current_event"
            assert current_event_data["description"] == "Current event description"
            assert len(current_event_data["choices"]) == 2
            assert current_event_data["choices"][0]["text"] == "Choice 1"
            assert current_event_data["items"] == ["sword"]
            assert current_event_data["location"] == "Current location"
            assert current_event_data["act"] == 0
            assert current_event_data["scene"] == 1

            # Check event history
            assert len(saved_data["event_history"]) == 2
            assert saved_data["event_history"][0]["id"] == "event_0"
            assert saved_data["event_history"][1]["id"] == "current_event"

            # Check action log
            assert len(saved_data["action_log"]) == 2
            assert saved_data["action_log"][0]["action_type"] == "start_game"
            assert saved_data["action_log"][1]["action_type"] == "make_choice"

            # Check game state
            game_state_data = saved_data["game_state"]
            assert game_state_data["inventory"] == ["potion"]
            assert game_state_data["act"] == 0
            assert game_state_data["scene"] == 1
            assert game_state_data["visited_events"] == ["event_0", "current_event"]
            assert game_state_data["flags"] == {"started": True}

        finally:
            if os.path.exists(filename):
                os.unlink(filename)

    def test_save_game_no_story_data(self, save_manager):
        """Test saving game with no story data."""
        success, message = save_manager.save_game(
            "test.json", None, GameState(), [], [], 0
        )

        assert success is False
        assert "No story data to save" in message

    def test_save_game_no_current_event(self, save_manager):
        """Test saving game with no current event."""
        story_data = StoryData(title="Test")
        game_state = GameState()  # No current event

        with tempfile.NamedTemporaryFile(mode="w", delete=False, suffix=".json") as f:
            filename = f.name

        try:
            success, message = save_manager.save_game(
                filename, story_data, game_state, [], [], 0
            )

            assert success is True

            with open(filename, "r") as f:
                saved_data = json.load(f)

            assert saved_data["current_event"] is None

        finally:
            if os.path.exists(filename):
                os.unlink(filename)

    @patch("builtins.open", side_effect=PermissionError("Permission denied"))
    def test_save_game_file_error(self, mock_file, save_manager):
        """Test saving game with file write error."""
        story_data = StoryData(title="Test")

        success, message = save_manager.save_game(
            "test.json", story_data, GameState(), [], [], 0
        )

        assert success is False
        assert "Error saving game" in message
        assert "Permission denied" in message

    def test_load_saved_game_complete(self, save_manager, sample_saved_game_data):
        """Test loading a complete saved game."""
        game_state, event_history, action_log, event_counter = (
            save_manager.load_saved_game(sample_saved_game_data)
        )

        # Check event counter
        assert event_counter == 5

        # Check game state
        assert game_state.inventory == ["potion"]
        assert game_state.act == 0
        assert game_state.scene == 1
        assert game_state.visited_events == ["event_0", "event_1"]
        assert game_state.flags == {"started": True}

        # Check current event (should be set to last event in history)
        assert game_state.current_event is not None
        assert game_state.current_event.id == "event_1"
        assert game_state.current_event.description == "You started the adventure"
        assert len(game_state.current_event.choices) == 2
        assert game_state.current_event.items == ["potion"]
        assert game_state.current_event.location == "Crossroads"

        # Check event history
        assert len(event_history) == 2
        assert event_history[0].id == "event_0"
        assert event_history[0].description == "The beginning"
        assert len(event_history[0].choices) == 2
        assert event_history[0].choices[0].text == "Start"
        assert event_history[0].items == []
        assert event_history[0].location == "Start"
        assert event_history[0].act == 0
        assert event_history[0].scene == 0

        assert event_history[1].id == "event_1"
        assert event_history[1].description == "You started the adventure"

        # Check action log
        assert len(action_log) == 2
        assert action_log[0].timestamp == 0
        assert action_log[0].action_type == "make_choice"
        assert action_log[0].details == "Made choice: Start"

        assert action_log[1].timestamp == 1
        assert action_log[1].action_type == "update_state"
        assert action_log[1].details == "Moved to event: event_1"

    def test_load_saved_game_minimal(self, save_manager):
        """Test loading saved game with minimal data."""
        minimal_save_data = {
            "event_counter": 1,
            "game_state": {},
            "event_history": [],
            "action_log": [],
        }

        game_state, event_history, action_log, event_counter = (
            save_manager.load_saved_game(minimal_save_data)
        )

        assert event_counter == 1
        assert game_state.inventory == []
        assert game_state.act == 0
        assert game_state.scene == 0
        assert game_state.visited_events == []
        assert game_state.flags == {}
        assert game_state.current_event is None
        assert event_history == []
        assert action_log == []

    def test_load_saved_game_no_current_event(self, save_manager):
        """Test loading saved game with empty event history."""
        save_data = {
            "event_counter": 0,
            "game_state": {"inventory": ["item"]},
            "event_history": [],
            "action_log": [],
        }

        game_state, event_history, action_log, event_counter = (
            save_manager.load_saved_game(save_data)
        )

        assert game_state.current_event is None
        assert game_state.inventory == ["item"]
        assert event_history == []

    def test_load_saved_game_choice_formats(self, save_manager):
        """Test loading saved game with different choice formats."""
        save_data = {
            "event_counter": 1,
            "game_state": {},
            "event_history": [
                {
                    "id": "test_event",
                    "description": "Test",
                    "choices": [
                        {"text": "Dict choice"},  # Dict format
                        "String choice",  # String format
                        {"text": ""},  # Empty text
                        {"not_text": "Invalid"},  # Invalid dict
                    ],
                    "items": [],
                    "location": "Test",
                    "act": 0,
                    "scene": 0,
                }
            ],
            "action_log": [],
        }

        game_state, event_history, action_log, event_counter = (
            save_manager.load_saved_game(save_data)
        )

        assert len(event_history) == 1
        event = event_history[0]

        # Should have processed valid choices
        assert len(event.choices) >= 2  # At least the valid ones
        assert any(choice.text == "Dict choice" for choice in event.choices)
        assert any(choice.text == "String choice" for choice in event.choices)

    def test_roundtrip_save_load(self, save_manager, sample_story_data):
        """Test complete save and load roundtrip."""
        # Create original data
        story_data = StoryData(
            title="Roundtrip Test",
            setting="Test World",
            acts=sample_story_data["acts"],
            npcs=sample_story_data["npcs"],
            items=sample_story_data["items"],
        )

        original_event = Event(
            "test_event",
            "Test description",
            [Choice("Test choice 1"), Choice("Test choice 2")],
            ["test_item"],
            "Test location",
            1,
            1,
        )

        original_game_state = GameState(
            current_event=original_event,
            inventory=["item1", "item2"],
            act=1,
            scene=1,
            visited_events=["event_0", "test_event"],
            flags={"test_flag": True, "level": 5},
        )

        original_event_history = [original_event]
        original_action_log = [ActionLogEntry(1, "test_action", "Test details")]
        original_event_counter = 5

        # Save the data
        with tempfile.NamedTemporaryFile(mode="w", delete=False, suffix=".json") as f:
            filename = f.name

        try:
            success, _ = save_manager.save_game(
                filename,
                story_data,
                original_game_state,
                original_event_history,
                original_action_log,
                original_event_counter,
            )
            assert success is True

            # Load the saved data
            with open(filename, "r") as f:
                saved_data = json.load(f)

            (
                loaded_game_state,
                loaded_event_history,
                loaded_action_log,
                loaded_event_counter,
            ) = save_manager.load_saved_game(saved_data)

            # Verify everything matches
            assert loaded_event_counter == original_event_counter

            assert loaded_game_state.inventory == original_game_state.inventory
            assert loaded_game_state.act == original_game_state.act
            assert loaded_game_state.scene == original_game_state.scene
            assert (
                loaded_game_state.visited_events == original_game_state.visited_events
            )
            assert loaded_game_state.flags == original_game_state.flags

            assert len(loaded_event_history) == len(original_event_history)
            loaded_event = loaded_event_history[0]
            assert loaded_event.id == original_event.id
            assert loaded_event.description == original_event.description
            assert len(loaded_event.choices) == len(original_event.choices)
            assert loaded_event.choices[0].text == original_event.choices[0].text
            assert loaded_event.items == original_event.items
            assert loaded_event.location == original_event.location
            assert loaded_event.act == original_event.act
            assert loaded_event.scene == original_event.scene

            assert len(loaded_action_log) == len(original_action_log)
            loaded_action = loaded_action_log[0]
            assert loaded_action.timestamp == original_action_log[0].timestamp
            assert loaded_action.action_type == original_action_log[0].action_type
            assert loaded_action.details == original_action_log[0].details

        finally:
            if os.path.exists(filename):
                os.unlink(filename)



"""Tests for main module and shell commands."""

import json
import os
import tempfile
from io import StringIO
from unittest.mock import MagicMock, Mock, patch

import pytest

from src.texventure.main import TexventureShell, main
from src.texventure.models import Choice, Event


class TestTexventureShell:
    """Test cases for TexventureShell."""

    def test_shell_initialization_minimal(self):
        """Test minimal shell initialization."""
        shell = TexventureShell()

        assert shell.console is not None
        assert shell.engine is not None
        assert shell.ui is not None
        assert shell.prompt == "(texventure) "
        assert shell.api_key is None
        assert shell.model == "gpt-3.5-turbo"

    def test_shell_initialization_with_params(self):
        """Test shell initialization with parameters."""
        shell = TexventureShell(api_key="test_key", model="gpt-4", use_llm=False)

        assert shell.api_key == "test_key"
        assert shell.model == "gpt-4"
        assert shell.engine.use_llm is False

    @patch.dict("os.environ", {})
    def test_shell_sets_api_key_env(self):
        """Test that shell sets API key in environment."""
        api_key = "test_api_key"
        shell = TexventureShell(api_key=api_key)

        assert os.environ.get("OPENAI_API_KEY") == api_key

    def test_load_story_new_game(self, sample_story_data):
        """Test loading a new story."""
        with tempfile.NamedTemporaryFile(mode="w", delete=False, suffix=".json") as f:
            json.dump(sample_story_data, f)
            filename = f.name

        try:
            shell = TexventureShell()

            with (
                patch.object(shell.ui, "render_title_screen") as mock_title,
                patch.object(shell.ui, "render_event") as mock_event,
                patch.object(shell.engine, "update_state") as mock_update,
            ):

                mock_event_obj = Event("test", "test event")
                mock_update.return_value = mock_event_obj

                shell.load_story(filename)

                assert shell.engine.story_data.title == "Test Adventure"
                assert shell.prompt == "(Test Adventure) "
                mock_title.assert_called_once()
                mock_update.assert_called_once_with("begin")
                mock_event.assert_called_once_with(mock_event_obj)

        finally:
            if os.path.exists(filename):
                os.unlink(filename)

    def test_load_story_saved_game(self, sample_saved_game_data):
        """Test loading a saved game."""
        with tempfile.NamedTemporaryFile(mode="w", delete=False, suffix=".json") as f:
            json.dump(sample_saved_game_data, f)
            filename = f.name

        try:
            shell = TexventureShell()

            with (
                patch.object(shell.ui, "render_title_screen") as mock_title,
                patch.object(shell.ui, "render_event") as mock_event,
                patch.object(shell.console, "print") as mock_print,
            ):

                shell.load_story(filename)

                assert shell.engine.story_data.title == "The mysteries of Mr Cat"
                mock_title.assert_called_once()
                mock_event.assert_called_once()

                # Check that saved game messages were printed
                print_calls = [call[0][0] for call in mock_print.call_args_list]
                assert any("Loading saved game" in str(call) for call in print_calls)

        finally:
            if os.path.exists(filename):
                os.unlink(filename)

    def test_load_story_file_not_found(self):
        """Test loading non-existent story file."""
        shell = TexventureShell()

        with (
            patch.object(shell.ui, "render_error") as mock_error,
            patch("sys.exit") as mock_exit,
        ):

            shell.load_story("nonexistent.json")

            mock_error.assert_called_once()
            mock_exit.assert_called_once_with(1)

    def test_load_story_invalid_json(self):
        """Test loading invalid JSON file."""
        with tempfile.NamedTemporaryFile(mode="w", delete=False, suffix=".json") as f:
            f.write("Invalid JSON content")
            filename = f.name

        try:
            shell = TexventureShell()

            with (
                patch.object(shell.ui, "render_error") as mock_error,
                patch("sys.exit") as mock_exit,
            ):

                shell.load_story(filename)

                mock_error.assert_called_once()
                error_message = mock_error.call_args[0][0]
                assert "Invalid JSON" in error_message
                mock_exit.assert_called_once_with(1)

        finally:
            if os.path.exists(filename):
                os.unlink(filename)

    def test_do_look_with_current_event(self, sample_story_data):
        """Test look command with current event."""
        shell = TexventureShell()
        shell.engine.load_story(sample_story_data)

        current_event = Event("test", "Test event")
        shell.engine.game_state.current_event = current_event

        with patch.object(shell.ui, "render_event") as mock_render:
            shell.do_look("")
            mock_render.assert_called_once_with(current_event)

    def test_do_look_no_current_event(self):
        """Test look command without current event."""
        shell = TexventureShell()

        with patch.object(shell.ui, "render_error") as mock_error:
            shell.do_look("")
            mock_error.assert_called_once_with("No current event to display.")

    def test_do_inventory(self, sample_story_data):
        """Test inventory command."""
        shell = TexventureShell()
        shell.engine.load_story(sample_story_data)
        shell.engine.game_state.inventory = ["sword", "potion"]

        with patch.object(shell.ui, "render_inventory") as mock_render:
            shell.do_inventory("")

            mock_render.assert_called_once()
            call_args = mock_render.call_args[0]
            assert call_args[0] == ["sword", "potion"]  # inventory
            assert call_args[1] == shell.engine.story_data.items  # items_data

    def test_do_take_success(self, sample_story_data):
        """Test successful take command."""
        shell = TexventureShell()
        shell.engine.load_story(sample_story_data)

        with (
            patch.object(shell.engine, "add_to_inventory") as mock_add,
            patch.object(shell.ui, "render_success") as mock_success,
        ):

            mock_add.return_value = (True, "You take the sword")

            shell.do_take("sword")

            mock_add.assert_called_once_with("sword")
            mock_success.assert_called_once_with("You take the sword")

    def test_do_take_failure(self, sample_story_data):
        """Test failed take command."""
        shell = TexventureShell()
        shell.engine.load_story(sample_story_data)

        with (
            patch.object(shell.engine, "add_to_inventory") as mock_add,
            patch.object(shell.ui, "render_error") as mock_error,
        ):

            mock_add.return_value = (False, "Item not found")

            shell.do_take("nonexistent")

            mock_add.assert_called_once_with("nonexistent")
            mock_error.assert_called_once_with("Item not found")

    def test_do_take_no_item(self):
        """Test take command without item name."""
        shell = TexventureShell()

        with patch.object(shell.ui, "render_error") as mock_error:
            shell.do_take("")
            mock_error.assert_called_once_with("Please specify an item to take.")

    def test_do_choice_success(self, sample_story_data):
        """Test successful choice command."""
        shell = TexventureShell()
        shell.engine.load_story(sample_story_data)

        new_event = Event("new", "New event")

        with (
            patch.object(shell.engine, "make_choice") as mock_choice,
            patch.object(shell.ui, "render_choice_feedback") as mock_feedback,
            patch.object(shell.ui, "render_event") as mock_event,
        ):

            mock_choice.return_value = (True, "You chose: Go north", new_event)

            shell.do_choice("1")

            mock_choice.assert_called_once_with(0)  # 1-indexed to 0-indexed
            mock_feedback.assert_called_once_with("Go north")  # Prefix removed
            mock_event.assert_called_once_with(new_event)

    def test_do_choice_failure(self):
        """Test failed choice command."""
        shell = TexventureShell()

        with (
            patch.object(shell.engine, "make_choice") as mock_choice,
            patch.object(shell.ui, "render_error") as mock_error,
        ):

            mock_choice.return_value = (False, "Invalid choice", None)

            shell.do_choice("10")

            mock_choice.assert_called_once_with(9)  # 1-indexed to 0-indexed
            mock_error.assert_called_once_with("Invalid choice")

    def test_do_choice_invalid_input(self):
        """Test choice command with invalid input."""
        shell = TexventureShell()

        with patch.object(shell.ui, "render_error") as mock_error:
            shell.do_choice("not_a_number")
            mock_error.assert_called_once_with("Please enter a valid choice number.")

    def test_do_save_with_filename(self, sample_story_data):
        """Test save command with filename."""
        shell = TexventureShell()
        shell.engine.load_story(sample_story_data)

        with (
            patch.object(shell.engine, "save_game") as mock_save,
            patch.object(shell.ui, "render_success") as mock_success,
            patch.object(shell.ui, "render_info") as mock_info,
        ):

            mock_save.return_value = (True, "Game saved to test.json")
            shell.engine.event_history = [Event("e1", "Event 1")]
            shell.engine.action_log = [Mock(), Mock()]

            shell.do_save("test.json")

            mock_save.assert_called_once_with("test.json")
            mock_success.assert_called_once_with("Game saved to test.json")
            mock_info.assert_called_once()

    def test_do_save_default_filename(self, sample_story_data):
        """Test save command with default filename."""
        shell = TexventureShell()
        shell.engine.load_story(sample_story_data)

        with patch.object(shell.engine, "save_game") as mock_save:
            mock_save.return_value = (True, "Game saved")

            shell.do_save("")

            expected_filename = "Test Adventure-savegame.json"
            mock_save.assert_called_once_with(expected_filename)

    def test_do_save_failure(self, sample_story_data):
        """Test failed save command."""
        shell = TexventureShell()
        shell.engine.load_story(sample_story_data)

        with (
            patch.object(shell.engine, "save_game") as mock_save,
            patch.object(shell.ui, "render_error") as mock_error,
        ):

            mock_save.return_value = (False, "Save failed")

            shell.do_save("test.json")

            mock_error.assert_called_once_with("Save failed")

    def test_do_history(self):
        """Test history command."""
        shell = TexventureShell()

        with patch.object(shell.ui, "render_action_history") as mock_render:
            shell.do_history("")
            mock_render.assert_called_once_with(shell.engine.action_log)

    def test_do_status(self, sample_story_data):
        """Test status command."""
        shell = TexventureShell()
        shell.engine.load_story(sample_story_data)

        with (
            patch.object(shell.ui, "render_status") as mock_status,
            patch.object(shell.console, "print") as mock_print,
        ):

            shell.do_status("")

            mock_status.assert_called_once_with(
                shell.engine.game_state,
                shell.engine.event_history,
                shell.engine.story_data,
            )

            # Check that LLM status was printed
            print_calls = [call[0][0] for call in mock_print.call_args_list]
            assert any("LLM Configuration" in str(call) for call in print_calls)

    def test_default_numeric_input(self, sample_story_data):
        """Test default handler with numeric input."""
        shell = TexventureShell()
        shell.engine.load_story(sample_story_data)

        with patch.object(shell, "do_choice") as mock_choice:
            shell.default("2")
            mock_choice.assert_called_once_with("2")

    def test_default_invalid_command(self):
        """Test default handler with invalid command."""
        shell = TexventureShell()

        with (
            patch.object(shell.ui, "render_error") as mock_error,
            patch.object(shell.ui, "render_info") as mock_info,
        ):

            shell.default("invalid_command")

            mock_error.assert_called_once()
            mock_info.assert_called_once()

    def test_do_exit(self):
        """Test exit command."""
        shell = TexventureShell()

        with patch.object(shell.ui, "render_goodbye") as mock_goodbye:
            result = shell.do_exit("")

            mock_goodbye.assert_called_once()
            assert result is True  # Should return True to exit


class TestMainFunction:
    """Test cases for main function."""

    @patch("src.texventure.main.TexventureShell")
    @patch("sys.argv", ["texventure", "test.json"])
    def test_main_basic(self, mock_shell_class):
        """Test basic main function execution."""
        mock_shell = Mock()
        mock_shell_class.return_value = mock_shell

        with patch("os.path.exists", return_value=True):
            main()

            mock_shell_class.assert_called_once()
            mock_shell.cmdloop.assert_called_once()

    @patch("sys.argv", ["texventure", "nonexistent.json"])
    @patch("sys.exit")
    def test_main_file_not_found(self, mock_exit):
        """Test main function with non-existent file."""
        with (
            patch("os.path.exists", return_value=False),
            patch("rich.console.Console.print") as mock_print,
        ):

            main()

            mock_exit.assert_called_once_with(1)
            # Check that error was printed
            assert mock_print.called

    @patch("src.texventure.main.TexventureShell")
    @patch(
        "sys.argv",
        ["texventure", "test.json", "--api-key", "test_key", "--model", "gpt-4"],
    )
    def test_main_with_args(self, mock_shell_class):
        """Test main function with command line arguments."""
        mock_shell = Mock()
        mock_shell_class.return_value = mock_shell

        with patch("os.path.exists", return_value=True):
            main()

            # Check that shell was created with correct arguments
            call_args = mock_shell_class.call_args[1]
            assert call_args["api_key"] == "test_key"
            assert call_args["model"] == "gpt-4"
            assert call_args["use_llm"] is True

    @patch("src.texventure.main.TexventureShell")
    @patch("sys.argv", ["texventure", "test.json", "--no-llm"])
    def test_main_no_llm(self, mock_shell_class):
        """Test main function with --no-llm flag."""
        mock_shell = Mock()
        mock_shell_class.return_value = mock_shell

        with patch("os.path.exists", return_value=True):
            main()

            call_args = mock_shell_class.call_args[1]
            assert call_args["use_llm"] is False

    @patch("src.texventure.main.TexventureShell")
    @patch("sys.argv", ["texventure", "test.json"])
    @patch.dict("os.environ", {}, clear=True)
    def test_main_llm_no_api_key_warning(self, mock_shell_class):
        """Test main function with LLM enabled but no API key."""
        mock_shell = Mock()
        mock_shell_class.return_value = mock_shell

        with (
            patch("os.path.exists", return_value=True),
            patch("rich.console.Console.print") as mock_print,
        ):

            main()

            # Check that warning was printed
            print_calls = [call[0][0] for call in mock_print.call_args_list]
            assert any("Warning" in str(call) for call in print_calls)

    @patch("src.texventure.main.TexventureShell")
    @patch("sys.argv", ["texventure", "test.json"])
    def test_main_keyboard_interrupt(self, mock_shell_class):
        """Test main function with keyboard interrupt."""
        mock_shell = Mock()
        mock_shell.cmdloop.side_effect = KeyboardInterrupt()
        mock_shell_class.return_value = mock_shell

        with (
            patch("os.path.exists", return_value=True),
            patch("rich.console.Console.print") as mock_print,
        ):

            main()

            # Check that goodbye was printed
            print_calls = [call[0][0] for call in mock_print.call_args_list]
            assert any("Goodbye" in str(call) for call in print_calls)



"""Event generation strategies for Texventure."""

import json
import random
from abc import ABC, abstractmethod
from typing import List, Tuple

from .generator import llm_call
from .models import Choice, Event


class EventGenerator(ABC):
    """Abstract base class for event generation strategies."""

    @abstractmethod
    def generate_event(self, context: dict, choice_text: str, event_id: str) -> Event:
        """Generate an event based on context and choice."""
        pass


class LLMEventGenerator(EventGenerator):
    """Generates events using LLM."""

    def __init__(self, api_key: str, model: str = "gpt-4.1-nano"):
        self.api_key = api_key
        self.model = model

    def generate_event(self, context: dict, choice_text: str, event_id: str) -> Event:
        """Generate event using LLM."""
        # Note: Don't check for game end here - that's handled in the engine
        # after the final trigger is activated

        context_str = self._build_context_string(context)
        scene_trigger = self._get_current_scene_trigger(context)

        prompt = self._create_continuation_prompt(
            context_str, choice_text, scene_trigger
        )

        try:
            response = llm_call(prompt, self.api_key, force_json=True)
            if response:
                return self._parse_llm_response(response, event_id, context)
        except Exception as e:
            print(f"Error generating event with LLM: {e}")

        return self._create_fallback_event(event_id, choice_text, context)

    def _create_continuation_prompt(
        self, context: str, choice_text: str, scene_trigger: str
    ) -> str:
        """Create prompt for continuation event."""
        trigger_instruction = ""
        if scene_trigger:
            trigger_instruction = f"\nSCENE TRIGGER: {scene_trigger}\nInclude the scene trigger as one of the player choices (rewritten to fit the context)."

        return f"""You are generating the next event in an interactive text adventure game.

STORY CONTEXT:
{context}

PLAYER'S PREVIOUS CHOICE: {choice_text}{trigger_instruction}

Generate the next event. You must respond with valid JSON in this exact format:
{{
    "description": "A vivid description of what happens next (2-3 sentences)",
    "location": "The current location name",
    "choices": [
        {{"text": "First choice option"}},
        {{"text": "Second choice option"}},
        {{"text": "Third choice option"}},
        {{"text": "Fourth choice option"}}
    ],
    "trigger_idx": 3
}}

Requirements:
1. Event logically follows from the player's choice
2. Fits the story's setting and tone
3. Advances the narrative meaningfully
4. Provides impactful choices
5. Includes engaging NPC dialogue when appropriate

Respond only with valid JSON in the specified format."""

    def _build_context_string(self, context: dict) -> str:
        """Build context string for LLM prompt."""
        context_parts = []
        story_data = context.get("story_data")
        game_state = context.get("game_state")

        # Add story basics
        if story_data:
            context_parts.extend(
                [
                    f"Title: {story_data.title}",
                    f"Setting: {story_data.setting}",
                    f"Outline: {story_data.outline}",
                    f"Style/Genre: {story_data.style_genre}",
                ]
            )

        # Add current act/scene info
        if (
            story_data
            and game_state
            and self._is_valid_act_scene(story_data, game_state)
        ):
            current_act = story_data.acts[game_state.act]
            context_parts.extend(
                [
                    f"Current Act: {current_act.get('name', f'Act {game_state.act + 1}')}",
                    f"Act Description: {current_act.get('description', '')}",
                ]
            )

            # Add current scene info
            scenes = current_act.get("scenes", [])
            if scenes and game_state.scene < len(scenes):
                current_scene = scenes[game_state.scene]
                context_parts.extend(
                    [
                        f"Current Scene: {current_scene.get('name', f'Scene {game_state.scene + 1}')}",
                        f"Scene Description: {current_scene.get('description', '')}",
                        f"Location: {current_scene.get('location', '')}",
                    ]
                )

        # Add NPCs context
        if story_data and story_data.npcs:
            context_parts.append("Available NPCs:")
            for npc in story_data.npcs[:3]:  # Limit to first 3 NPCs
                context_parts.append(
                    f"- {npc.get('name', 'Unknown')}: {npc.get('description', '')}"
                )

        # Add event context
        self._add_event_context(context_parts, context)

        # Add inventory
        if game_state and game_state.inventory:
            context_parts.append(f"Player Inventory: {', '.join(game_state.inventory)}")

        return "\n".join(context_parts)

    def _is_valid_act_scene(self, story_data, game_state) -> bool:
        """Check if the current act and scene indices are valid."""
        return (
            hasattr(story_data, "acts")
            and story_data.acts
            and game_state.act < len(story_data.acts)
        )

    def _add_event_context(self, context_parts: List[str], context: dict) -> None:
        """Add current and previous event context to context parts."""
        current_event = context.get("current_event")
        if current_event:
            context_parts.extend(
                [
                    f"Current Event: {current_event.description}",
                    f"Current Location: {current_event.location}",
                ]
            )
            if current_event.items:
                context_parts.append(
                    f"Available Items: {', '.join(current_event.items)}"
                )

        previous_event = context.get("previous_event")
        if previous_event:
            context_parts.extend(
                [
                    f"Previous Event: {previous_event.description}",
                    f"Previous Location: {previous_event.location}",
                ]
            )

    def _get_current_scene_trigger(self, context: dict) -> str:
        """Get the trigger for the current scene."""
        try:
            story_data = context.get("story_data")
            game_state = context.get("game_state")

            if not (
                story_data
                and game_state
                and self._is_valid_act_scene(story_data, game_state)
            ):
                return ""

            current_act = story_data.acts[game_state.act]
            scenes = current_act.get("scenes", [])

            if scenes and game_state.scene < len(scenes):
                return scenes[game_state.scene].get("trigger", "")

            return ""
        except (AttributeError, IndexError, KeyError):
            return ""

    def _shuffle_choices_with_trigger(
        self, choices: List[Choice], trigger_idx: int
    ) -> Tuple[List[Choice], int]:
        """Shuffle choices while maintaining trigger index tracking."""
        if len(choices) <= 1:
            return choices, trigger_idx

        # Create choice-trigger pairs
        choice_items = [(choice, i == trigger_idx) for i, choice in enumerate(choices)]
        random.shuffle(choice_items)

        # Rebuild choices and find new trigger index
        new_choices = []
        new_trigger_idx = -1
        for i, (choice, is_trigger) in enumerate(choice_items):
            new_choices.append(choice)
            if is_trigger:
                new_trigger_idx = i

        return new_choices, new_trigger_idx

    def _ensure_scene_trigger_in_choices(
        self, choices: List[Choice], scene_trigger: str, trigger_idx: int
    ) -> Tuple[List[Choice], int]:
        """Ensure scene trigger is included in choices if it exists."""
        if not scene_trigger or not scene_trigger.strip():
            return choices, trigger_idx

        # Check if trigger already exists in choices
        for i, choice in enumerate(choices):
            if (
                scene_trigger.lower() in choice.text.lower()
                or choice.text.lower() in scene_trigger.lower()
                or trigger_idx == i
            ):
                return choices, i

        # Add or replace with trigger
        if choices:
            choices[-1] = Choice(scene_trigger)
            return choices, len(choices) - 1
        else:
            return [Choice(scene_trigger)], 0

    def _parse_llm_response(
        self, llm_response: str, event_id: str, context: dict
    ) -> Event:
        """Parse LLM response and create Event object."""
        try:
            event_data = json.loads(llm_response.strip())

            # Validate and extract basic fields
            description = event_data.get("description", "")
            if not description:
                raise ValueError("Missing description in LLM response")

            location = event_data.get("location", "Unknown location")
            choices = self._parse_choices(event_data.get("choices", []))

            if not choices:
                raise ValueError("No valid choices in LLM response")

            # Handle scene trigger
            scene_trigger = self._get_current_scene_trigger(context)
            trigger_idx = event_data.get("trigger_idx", -1)
            choices, trigger_idx = self._ensure_scene_trigger_in_choices(
                choices, scene_trigger, trigger_idx
            )

            # Shuffle choices while preserving trigger index
            choices, trigger_idx = self._shuffle_choices_with_trigger(
                choices, trigger_idx
            )

            game_state = context.get("game_state")
            return Event(
                id=event_id,
                description=description,
                choices=choices,
                location=location,
                act=game_state.act if game_state else 0,
                scene=game_state.scene if game_state else 0,
                trigger_choice_idx=trigger_idx,
            )

        except (json.JSONDecodeError, KeyError, ValueError) as e:
            print(f"Error parsing LLM response: {e}")
            print(f"Raw response: {llm_response}")
            return self._create_fallback_event(event_id, "continue", context)

    def _parse_choices(self, choice_data: list) -> List[Choice]:
        """Parse choices from LLM response data."""
        choices = []
        for choice_item in choice_data:
            choice_text = ""
            if isinstance(choice_item, dict):
                choice_text = choice_item.get("text", "")
            elif isinstance(choice_item, str):
                choice_text = choice_item

            if choice_text:
                choices.append(Choice(choice_text))

        return choices

    def _create_fallback_event(
        self, event_id: str, choice_text: str, context: dict
    ) -> Event:
        """Create a fallback event when LLM fails."""
        game_state = context.get("game_state")

        # Default choices
        choices = [
            Choice("Explore the area"),
            Choice("Look around carefully"),
            Choice("Rest and think"),
            Choice("Move forward"),
        ]

        # Handle scene trigger
        scene_trigger = self._get_current_scene_trigger(context)
        choices, trigger_idx = self._ensure_scene_trigger_in_choices(
            choices, scene_trigger, -1
        )
        choices, trigger_idx = self._shuffle_choices_with_trigger(choices, trigger_idx)

        return Event(
            id=event_id,
            description="The adventure continues as you find yourself in a new situation...",
            choices=choices,
            location="Unknown location",
            act=game_state.act if game_state else 0,
            scene=game_state.scene if game_state else 0,
            trigger_choice_idx=trigger_idx,
        )


class TemplateEventGenerator(EventGenerator):
    """Generates events using predefined templates."""

    def generate_event(self, context: dict, choice_text: str, event_id: str) -> Event:
        """Generate event using templates."""
        story_data = context.get("story_data")
        game_state = context.get("game_state")

        # Note: Don't check for game end here - that's handled in the engine
        # after the final trigger is activated

        if choice_text == "begin" and story_data and story_data.acts:
            description, location = self._get_opening_content(story_data)
        else:
            description = (
                f"You chose: {choice_text}. You find yourself in a new situation."
            )
            location = "Current location"

        choices = [
            Choice("Explore the area"),
            Choice("Look for items"),
            Choice("Rest and recover"),
            Choice("Continue forward"),
        ]

        return Event(
            id=event_id,
            description=description,
            choices=choices,
            location=location,
            act=game_state.act if game_state else 0,
            scene=game_state.scene if game_state else 0,
        )

    def _get_opening_content(self, story_data) -> Tuple[str, str]:
        """Get opening description and location from story data."""
        first_act = story_data.acts[0]
        first_scene = (
            first_act.get("scenes", [{}])[0] if first_act.get("scenes") else {}
        )

        description = (
            first_scene.get("description", "")
            or first_act.get("description", "")
            or "You begin your adventure..."
        )
        location = first_scene.get("location", "Unknown location")

        return description, location



"""Data models for Texventure game engine."""

from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional


@dataclass
class Choice:
    """Represents a choice in an event."""

    text: str
    consequences: Dict[str, Any] = field(default_factory=dict)
    requirements: Dict[str, Any] = field(default_factory=dict)


@dataclass
class Event:
    """Represents a game event."""

    id: str
    description: str
    choices: List[Choice] = field(default_factory=list)
    items: List[str] = field(default_factory=list)
    location: str = ""
    act: int = 0
    scene: int = 0
    trigger_choice_idx: int = -1


@dataclass
class GameState:
    """Represents the current game state."""

    current_event: Optional[Event] = None
    inventory: List[str] = field(default_factory=list)
    act: int = 0
    scene: int = 0
    visited_events: List[str] = field(default_factory=list)
    flags: Dict[str, Any] = field(default_factory=dict)


@dataclass
class ActionLogEntry:
    """Represents an action log entry."""

    timestamp: int
    action_type: str
    details: str


@dataclass
class StoryData:
    """Represents the loaded story data."""

    title: str = "Text Adventure"
    setting: str = "Unknown"
    outline: str = ""
    style_genre: str = ""
    acts: List[Dict[str, Any]] = field(default_factory=list)
    npcs: List[Dict[str, Any]] = field(default_factory=list)
    items: Dict[str, Dict[str, Any]] = field(default_factory=dict)



"""UI rendering components for Texventure."""

from typing import Any, Dict, List, Optional, Union

from rich import box
from rich.console import Console
from rich.panel import Panel
from rich.table import Table

from .models import ActionLogEntry, Event, GameState, StoryData


class UIRenderer:
    """Handles all UI rendering for the game."""

    def __init__(self, console: Optional[Console] = None) -> None:
        self.console = console or Console()

    def _create_panel(
        self,
        content: str,
        title: str,
        border_style: str = "bright_blue",
        box_style: box.Box = box.SIMPLE,
    ) -> Panel:
        """Create a standardized panel with consistent styling."""
        return Panel(
            content,
            title=title,
            border_style=border_style,
            box=box_style,
        )

    def _create_table(
        self,
        title: str,
        border_style: str = "bright_cyan",
        box_style: box.Box = box.ROUNDED,
    ) -> Table:
        """Create a standardized table with consistent styling."""
        return Table(
            title=title,
            box=box_style,
            border_style=border_style,
        )

    def render_title_screen(
        self, story_data: StoryData, first_scene_desc: str = ""
    ) -> None:
        """Render the title screen when the game starts."""
        content = (
            f"[bold cyan]{story_data.title}[/bold cyan]\n\n"
            f"Welcome to Texventure. Type help or ? to list commands.\n\n"
            f"[italic]{first_scene_desc}[/italic]"
        )
        title_panel = self._create_panel(
            content,
            "[bold yellow]🎮 TEXVENTURE[/bold yellow]",
            border_style="bright_blue",
            box_style=box.DOUBLE,
        )
        self.console.print(title_panel)

    def render_event(self, event: Event) -> None:
        """Render the current event."""
        description_panel = self._create_panel(
            event.description,
            "[bold cyan]🎭 Event[/bold cyan]",
            border_style="cyan",
        )
        self.console.print(description_panel)
        self.render_choices(event.choices)

    def _extract_choice_text(self, choice: Union[Dict[str, Any], Any]) -> str:
        """Extract choice text from various choice formats."""
        if hasattr(choice, "text"):
            return choice.text
        if isinstance(choice, dict):
            return choice.get("text", "Unknown choice")
        return str(choice)

    def render_choices(self, choices: List[Any]) -> None:
        """Render available choices."""
        if not choices:
            no_choices_panel = self._create_panel(
                "[italic red]No choices available - this might be the end of the adventure[/italic red]",
                "",
                border_style="red",
            )
            self.console.print(no_choices_panel)
            return

        choice_table = Table(
            title="[bold bright_cyan]⚡ What do you want to do?[/bold bright_cyan]",
            box=box.SIMPLE_HEAD,
            show_header=False,
            min_width=40,
            border_style="bright_cyan",
        )

        for i, choice in enumerate(choices, 1):
            choice_text = self._extract_choice_text(choice)
            choice_table.add_row(
                f"[bold cyan]{i}.[/bold cyan]", f"[white]{choice_text}[/white]"
            )

        self.console.print(choice_table)

    def render_inventory(
        self, inventory: List[str], items_data: Dict[str, Any]
    ) -> None:
        """Render the player's inventory."""
        if not inventory:
            empty_panel = self._create_panel(
                "[italic bright_black]Your inventory is empty.[/italic bright_black]",
                "[bold yellow]🎒 Inventory[/bold yellow]",
                border_style="yellow",
            )
            self.console.print(empty_panel)
            return

        inv_table = self._create_table(
            "[bold yellow]🎒 Your Inventory[/bold yellow]",
            border_style="yellow",
        )
        inv_table.add_column("Item", style="cyan", no_wrap=True)
        inv_table.add_column("Description", style="white")

        for item_id in inventory:
            item = items_data.get(item_id, {})
            item_name = item.get("name", item_id)
            item_desc = item.get("description", "No description")
            inv_table.add_row(f"✨ {item_name}", item_desc)

        self.console.print(inv_table)

    def render_action_history(
        self, action_log: List[ActionLogEntry], limit: int = 10
    ) -> None:
        """Render recent action history."""
        if not action_log:
            self.console.print("[italic]No history available yet.[/italic]")
            return

        recent_actions = action_log[-limit:]
        action_table = self._create_table(
            "[bold yellow]📝 Recent Actions[/bold yellow]",
            border_style="yellow",
        )
        action_table.add_column("Step", style="yellow", width=8)
        action_table.add_column("Action", style="cyan")
        action_table.add_column("Details", style="white")

        for action in recent_actions:
            action_table.add_row(
                str(action.timestamp),
                action.action_type.replace("_", " ").title(),
                action.details,
            )

        self.console.print(action_table)

        if len(action_log) > limit:
            self.console.print(
                f"[italic]... and {len(action_log) - limit} more actions[/italic]"
            )

    def _render_current_scene(
        self, game_state: GameState, story_data: StoryData
    ) -> None:
        """Render current scene information if available."""
        if not (
            story_data and story_data.acts and game_state.act < len(story_data.acts)
        ):
            return

        current_act = story_data.acts[game_state.act]
        scenes = current_act.get("scenes", [])

        if not (scenes and game_state.scene < len(scenes)):
            return

        current_scene = scenes[game_state.scene]
        scene_description = current_scene.get("description", "")
        scene_name = current_scene.get("name", f"Scene {game_state.scene + 1}")

        if scene_description:
            scene_panel = self._create_panel(
                f"[bold yellow]{scene_name}[/bold yellow]\n\n{scene_description}",
                "[bold green]📍 Current Scene[/bold green]",
                border_style="green",
                box_style=box.ROUNDED,
            )
            self.console.print("\n")
            self.console.print(scene_panel)

    def render_status(
        self,
        game_state: GameState,
        event_history: List[Event],
        story_data: Optional[StoryData] = None,
    ) -> None:
        """Render current game status."""
        status_table = self._create_table(
            "[bold green]🎮 Game Status[/bold green]",
            border_style="green",
        )
        status_table.add_column("Metric", style="cyan", width=20)
        status_table.add_column("Value", style="white")

        current_location = (
            game_state.current_event.location if game_state.current_event else "Unknown"
        )

        status_data = [
            ("Events Visited", str(len(event_history))),
            ("Items in Inventory", str(len(game_state.inventory))),
            ("Current Act", str(game_state.act + 1)),
            ("Current Scene", str(game_state.scene + 1)),
            ("Current Location", current_location),
        ]

        for metric, value in status_data:
            status_table.add_row(metric, value)

        self.console.print(status_table)

        if story_data:
            self._render_current_scene(game_state, story_data)

    def render_choice_feedback(self, choice_text: str) -> None:
        """Render feedback when a choice is made."""
        choice_panel = self._create_panel(
            f"[bold cyan]> {choice_text}[/bold cyan]",
            "",
            border_style="bright_cyan",
        )
        self.console.print(choice_panel)

    def render_message(self, message: str, message_type: str = "info") -> None:
        """Render a message with the specified type (error, success, info)."""
        style_map = {
            "error": ("[bold red]✗", "[/bold red]"),
            "success": ("[bold green]✓", "[/bold green]"),
            "info": ("[cyan]", "[/cyan]"),
        }

        prefix, suffix = style_map.get(message_type, style_map["info"])
        self.console.print(f"{prefix} {message}{suffix}")

    def render_error(self, message: str) -> None:
        """Render an error message."""
        self.render_message(message, "error")

    def render_success(self, message: str) -> None:
        """Render a success message."""
        self.render_message(message, "success")

    def render_info(self, message: str) -> None:
        """Render an info message."""
        self.render_message(message, "info")

    def render_goodbye(self) -> None:
        """Render the goodbye screen."""
        goodbye_panel = self._create_panel(
            "[bold yellow]🌟 Goodbye! 🌟[/bold yellow]\n\n[italic]Thanks for playing Texventure![/italic]",
            "",
            border_style="yellow",
            box_style=box.DOUBLE,
        )
        self.console.print(goodbye_panel)



"""Save and load management for Texventure."""

import json
from typing import List, Tuple

from .models import ActionLogEntry, Choice, Event, GameState, StoryData


class SaveLoadManager:
    """Handles saving and loading game states."""

    @staticmethod
    def save_game(
        filename: str,
        story_data: StoryData,
        game_state: GameState,
        event_history: List[Event],
        action_log: List[ActionLogEntry],
        event_counter: int,
    ) -> Tuple[bool, str]:
        """Save the current game state in a format compatible with story loading."""
        if not story_data:
            return False, "No story data to save"

        # Prepare current event data
        current_event_data = None
        if game_state.current_event:
            current_event_data = {
                "id": game_state.current_event.id,
                "description": game_state.current_event.description,
                "choices": [
                    {"text": choice.text} for choice in game_state.current_event.choices
                ],
                "items": game_state.current_event.items,
                "location": game_state.current_event.location,
                "act": game_state.current_event.act,
                "scene": game_state.current_event.scene,
            }

        save_data = {
            # Core story data (compatible with original format)
            "title": story_data.title,
            "setting": story_data.setting,
            "outline": story_data.outline,
            "acts": story_data.acts,
            "npcs": story_data.npcs,
            "items": story_data.items,
            # Save game specific data
            "is_saved_game": True,
            "saved_at": event_counter,
            "current_event": current_event_data,
            "event_history": [
                {
                    "id": event.id,
                    "description": event.description,
                    "choices": [{"text": choice.text} for choice in event.choices],
                    "items": event.items,
                    "location": event.location,
                    "act": event.act,
                    "scene": event.scene,
                }
                for event in event_history
            ],
            "action_log": [
                {
                    "timestamp": entry.timestamp,
                    "action_type": entry.action_type,
                    "details": entry.details,
                }
                for entry in action_log
            ],
            "event_counter": event_counter,
            "game_state": {
                "inventory": game_state.inventory,
                "act": game_state.act,
                "scene": game_state.scene,
                "visited_events": game_state.visited_events,
                "flags": game_state.flags,
            },
        }

        try:
            with open(filename, "w") as f:
                json.dump(save_data, f, indent=2)
            return True, f"Game saved to {filename}"
        except Exception as e:
            return False, f"Error saving game: {e}"

    @staticmethod
    def load_saved_game(
        save_data: dict,
    ) -> Tuple[GameState, List[Event], List[ActionLogEntry], int]:
        """Load a saved game state and return components."""
        # Restore basic counters
        event_counter = save_data.get("event_counter", 0)

        # Restore game state
        saved_state = save_data.get("game_state", {})
        game_state = GameState(
            inventory=saved_state.get("inventory", []),
            act=saved_state.get("act", 0),
            scene=saved_state.get("scene", 0),
            visited_events=saved_state.get("visited_events", []),
            flags=saved_state.get("flags", {}),
        )

        # Restore event history
        event_history_data = save_data.get("event_history", [])
        event_history = []
        for event_data in event_history_data:
            choices = []
            for choice_data in event_data.get("choices", []):
                choice_text = (
                    choice_data.get("text", "")
                    if isinstance(choice_data, dict)
                    else choice_data
                )
                choices.append(Choice(choice_text))

            event = Event(
                id=event_data.get("id", ""),
                description=event_data.get("description", ""),
                choices=choices,
                items=event_data.get("items", []),
                location=event_data.get("location", ""),
                act=event_data.get("act", 0),
                scene=event_data.get("scene", 0),
            )
            event_history.append(event)

        # Set current event to the last one in history
        if event_history:
            game_state.current_event = event_history[-1]

        # Restore action log
        action_log_data = save_data.get("action_log", [])
        action_log = []
        for action_data in action_log_data:
            entry = ActionLogEntry(
                timestamp=action_data.get("timestamp", 0),
                action_type=action_data.get("action_type", ""),
                details=action_data.get("details", ""),
            )
            action_log.append(entry)

        return game_state, event_history, action_log, event_counter



"""Tests for game engine."""

from unittest.mock import Mock, patch

import pytest

from src.texventure.engine import GameEngine
from src.texventure.event_generators import (LLMEventGenerator,
                                             TemplateEventGenerator)
from src.texventure.models import (ActionLogEntry, Choice, Event, GameState,
                                   StoryData)


class TestGameEngine:
    """Test cases for GameEngine."""

    def test_game_engine_initialization_default(self):
        """Test default game engine initialization."""
        engine = GameEngine()

        assert engine.story_data is None
        assert isinstance(engine.game_state, GameState)
        assert engine.event_history == []
        assert engine.action_log == []
        assert engine.event_counter == 0
        assert engine.use_llm is True
        assert engine.model == "gpt-3.5-turbo"
        assert isinstance(
            engine.event_generator, (LLMEventGenerator, TemplateEventGenerator)
        )

    def test_game_engine_initialization_no_llm(self):
        """Test game engine initialization without LLM."""
        engine = GameEngine(use_llm=False)

        assert engine.use_llm is False
        assert isinstance(engine.event_generator, TemplateEventGenerator)

    def test_game_engine_initialization_custom_model(self):
        """Test game engine initialization with custom model."""
        engine = GameEngine(use_llm=True, model="gpt-4")

        assert engine.model == "gpt-4"

    @patch.dict("os.environ", {"OPENAI_API_KEY": "test_key"})
    def test_game_engine_llm_with_api_key(self):
        """Test game engine with LLM when API key is available."""
        engine = GameEngine(use_llm=True)
        assert isinstance(engine.event_generator, LLMEventGenerator)

    @patch.dict("os.environ", {}, clear=True)
    def test_game_engine_llm_without_api_key(self):
        """Test game engine with LLM when API key is not available."""
        engine = GameEngine(use_llm=True)
        # Should fall back to template generator
        assert isinstance(engine.event_generator, TemplateEventGenerator)

    def test_load_story_new_game(self, sample_story_data):
        """Test loading a new story."""
        engine = GameEngine()
        engine.load_story(sample_story_data)

        assert engine.story_data.title == "Test Adventure"
        assert engine.story_data.setting == "A test world"
        assert engine.story_data.outline == "A simple test story"
        assert len(engine.story_data.acts) == 2
        assert len(engine.story_data.npcs) == 1
        assert len(engine.story_data.items) == 2

        assert engine.game_state.act == 0
        assert engine.game_state.scene == 0
        assert engine.event_counter == 0

    def test_load_story_saved_game(self, sample_saved_game_data):
        """Test loading a saved game."""
        engine = GameEngine()
        engine.load_story(sample_saved_game_data)

        assert engine.story_data.title == "The mysteries of Mr Cat"
        assert engine.event_counter == 5
        assert engine.game_state.act == 0
        assert engine.game_state.scene == 1
        assert engine.game_state.inventory == ["potion"]
        assert len(engine.event_history) == 2
        assert len(engine.action_log) == 2
        assert engine.game_state.current_event is not None

    def test_log_action(self):
        """Test logging actions."""
        engine = GameEngine()
        engine.event_counter = 5

        engine.log_action("test_action", "Test details")

        assert len(engine.action_log) == 1
        action = engine.action_log[0]
        assert action.timestamp == 5
        assert action.action_type == "test_action"
        assert action.details == "Test details"

    def test_generate_event(self, sample_story_data):
        """Test event generation."""
        engine = GameEngine(use_llm=False)  # Use template to avoid LLM calls
        engine.load_story(sample_story_data)

        event = engine.generate_event("begin")

        assert event.id == "event_0"
        assert isinstance(event.description, str)
        assert len(event.description) > 0
        assert len(event.choices) == 4
        assert isinstance(event.location, str)

    def test_generate_event_no_story_data(self):
        """Test event generation without story data."""
        engine = GameEngine()

        with pytest.raises(ValueError, match="No story data loaded"):
            engine.generate_event("begin")

    def test_build_generation_context(self, sample_story_data):
        """Test building generation context."""
        engine = GameEngine()
        engine.load_story(sample_story_data)

        # Add some state
        current_event = Event("current", "Current event")
        engine.game_state.current_event = current_event

        previous_event = Event("previous", "Previous event")
        engine.event_history = [previous_event, current_event]

        context = engine._build_generation_context()

        assert context["story_data"] == engine.story_data
        assert context["game_state"] == engine.game_state
        assert context["current_event"] == current_event
        assert context["previous_event"] == previous_event

    def test_build_generation_context_no_previous(self, sample_story_data):
        """Test building generation context without previous event."""
        engine = GameEngine()
        engine.load_story(sample_story_data)

        current_event = Event("current", "Current event")
        engine.game_state.current_event = current_event
        engine.event_history = [current_event]  # Only one event

        context = engine._build_generation_context()

        assert "previous_event" not in context

    def test_update_state(self, sample_story_data):
        """Test updating game state."""
        engine = GameEngine(use_llm=False)
        engine.load_story(sample_story_data)

        initial_counter = engine.event_counter
        initial_log_count = len(engine.action_log)

        new_event = engine.update_state("test choice")

        assert engine.event_counter == initial_counter + 1
        assert engine.game_state.current_event == new_event
        assert new_event in engine.event_history
        assert new_event.id in engine.game_state.visited_events
        assert len(engine.action_log) == initial_log_count + 2  # choice + update

    def test_make_choice_valid(self, sample_story_data):
        """Test making a valid choice."""
        engine = GameEngine(use_llm=False)
        engine.load_story(sample_story_data)

        # Set up current event
        choices = [Choice("Choice 1"), Choice("Choice 2"), Choice("Choice 3")]
        current_event = Event("current", "Current event", choices=choices)
        engine.game_state.current_event = current_event

        success, message, new_event = engine.make_choice(1)  # Second choice (0-indexed)

        assert success is True
        assert "Choice 2" in message
        assert new_event is not None
        assert new_event != current_event

    def test_make_choice_invalid_index(self, sample_story_data):
        """Test making a choice with invalid index."""
        engine = GameEngine(use_llm=False)
        engine.load_story(sample_story_data)

        choices = [Choice("Choice 1"), Choice("Choice 2")]
        current_event = Event("current", "Current event", choices=choices)
        engine.game_state.current_event = current_event

        success, message, new_event = engine.make_choice(5)  # Invalid index

        assert success is False
        assert "Invalid choice" in message
        assert new_event is None

    def test_make_choice_no_current_event(self):
        """Test making a choice without current event."""
        engine = GameEngine()

        success, message, new_event = engine.make_choice(0)

        assert success is False
        assert "No current event" in message
        assert new_event is None

    def test_add_to_inventory_success(self, sample_story_data):
        """Test successfully adding item to inventory."""
        engine = GameEngine()
        engine.load_story(sample_story_data)

        # Set up current event with item
        current_event = Event("current", "Current event", items=["sword"])
        engine.game_state.current_event = current_event

        success, message = engine.add_to_inventory("sword")

        assert success is True
        assert "Magic Sword" in message  # Item name from story data
        assert "sword" in engine.game_state.inventory
        assert "sword" not in current_event.items

    def test_add_to_inventory_item_not_found(self, sample_story_data):
        """Test adding non-existent item to inventory."""
        engine = GameEngine()
        engine.load_story(sample_story_data)

        current_event = Event("current", "Current event")
        engine.game_state.current_event = current_event

        success, message = engine.add_to_inventory("nonexistent")

        assert success is False
        assert "not found" in message

    def test_add_to_inventory_already_have(self, sample_story_data):
        """Test adding item already in inventory."""
        engine = GameEngine()
        engine.load_story(sample_story_data)

        current_event = Event("current", "Current event", items=["sword"])
        engine.game_state.current_event = current_event
        engine.game_state.inventory = ["sword"]  # Already have it

        success, message = engine.add_to_inventory("sword")

        assert success is False
        assert "already have" in message

    def test_add_to_inventory_not_available_here(self, sample_story_data):
        """Test adding item not available in current event."""
        engine = GameEngine()
        engine.load_story(sample_story_data)

        current_event = Event("current", "Current event", items=[])  # No items
        engine.game_state.current_event = current_event

        success, message = engine.add_to_inventory("sword")

        assert success is False
        assert "not available here" in message

    def test_add_to_inventory_no_items_data(self):
        """Test adding item when no items data available."""
        engine = GameEngine()
        story_data = {"title": "Test", "items": {}}
        engine.load_story(story_data)

        success, message = engine.add_to_inventory("sword")

        assert success is False
        assert "No items available" in message

    def test_save_game(self, sample_story_data):
        """Test saving game."""
        engine = GameEngine()
        engine.load_story(sample_story_data)

        # Add some state
        current_event = Event("current", "Current event")
        engine.game_state.current_event = current_event
        engine.event_history.append(current_event)
        engine.log_action("test", "test action")

        with patch.object(engine.save_manager, "save_game") as mock_save:
            mock_save.return_value = (True, "Saved successfully")

            success, message = engine.save_game("test.json")

            assert success is True
            assert "Saved successfully" in message
            mock_save.assert_called_once_with(
                "test.json",
                engine.story_data,
                engine.game_state,
                engine.event_history,
                engine.action_log,
                engine.event_counter,
            )

            # Check that save action was logged
            assert any(
                action.action_type == "save_game" for action in engine.action_log
            )

    def test_save_game_failure(self, sample_story_data):
        """Test saving game failure."""
        engine = GameEngine()
        engine.load_story(sample_story_data)

        with patch.object(engine.save_manager, "save_game") as mock_save:
            mock_save.return_value = (False, "Save failed")

            success, message = engine.save_game("test.json")

            assert success is False
            assert "Save failed" in message

            # Check that save action was NOT logged on failure
            assert not any(
                action.action_type == "save_game" for action in engine.action_log
            )

    def test_integration_full_game_flow(self, sample_story_data):
        """Test full game flow integration."""
        engine = GameEngine(use_llm=False)

        # Load story
        engine.load_story(sample_story_data)
        assert engine.story_data is not None

        # Generate first event
        first_event = engine.update_state("begin")
        assert first_event.id == "event_1"  # Counter incremented
        assert engine.game_state.current_event == first_event
        assert len(engine.event_history) == 1

        # Make a choice
        success, message, second_event = engine.make_choice(0)
        assert success is True
        assert second_event != first_event
        assert len(engine.event_history) == 2

        # Add item if available
        if first_event.items:
            item_id = first_event.items[0]
            # Put item back for testing
            first_event.items.append(item_id)
            success, message = engine.add_to_inventory(item_id)
            assert success is True
            assert item_id in engine.game_state.inventory

        # Check action log has entries
        assert len(engine.action_log) > 0

        # Verify counters
        assert engine.event_counter == 2

    def test_event_generator_assignment(self):
        """Test that correct event generator is assigned based on configuration."""
        # Test without LLM
        engine_no_llm = GameEngine(use_llm=False)
        assert isinstance(engine_no_llm.event_generator, TemplateEventGenerator)

        # Test with LLM but no API key
        with patch.dict("os.environ", {}, clear=True):
            engine_llm_no_key = GameEngine(use_llm=True)
            assert isinstance(engine_llm_no_key.event_generator, TemplateEventGenerator)

        # Test with LLM and API key
        with patch.dict("os.environ", {"OPENAI_API_KEY": "test_key"}):
            engine_llm_with_key = GameEngine(use_llm=True)
            assert isinstance(engine_llm_with_key.event_generator, LLMEventGenerator)

    def test_event_generation_delegation(self, sample_story_data):
        """Test that event generation is properly delegated to event generator."""
        engine = GameEngine(use_llm=False)
        engine.load_story(sample_story_data)

        # Mock the event generator
        mock_generator = Mock()
        mock_event = Event("mock_event", "Mock description")
        mock_generator.generate_event.return_value = mock_event
        engine.event_generator = mock_generator

        result = engine.generate_event("test choice")

        assert result == mock_event
        mock_generator.generate_event.assert_called_once()

        # Check the context passed to generator
        call_args = mock_generator.generate_event.call_args
        context, choice_text, event_id = call_args[0]

        assert context["story_data"] == engine.story_data
        assert context["game_state"] == engine.game_state
        assert choice_text == "test choice"
        assert event_id == "event_0"



"""Tests for data models."""

from dataclasses import FrozenInstanceError

import pytest

from src.texventure.models import (ActionLogEntry, Choice, Event, GameState,
                                   StoryData)


class TestChoice:
    """Test cases for Choice model."""

    def test_choice_creation(self):
        """Test basic choice creation."""
        choice = Choice("Test choice")
        assert choice.text == "Test choice"
        assert choice.consequences == {}
        assert choice.requirements == {}

    def test_choice_with_consequences(self):
        """Test choice with consequences."""
        consequences = {"health": -10, "experience": 5}
        choice = Choice("Dangerous choice", consequences=consequences)
        assert choice.text == "Dangerous choice"
        assert choice.consequences == consequences

    def test_choice_with_requirements(self):
        """Test choice with requirements."""
        requirements = {"level": 5, "item": "sword"}
        choice = Choice("Advanced choice", requirements=requirements)
        assert choice.text == "Advanced choice"
        assert choice.requirements == requirements

    def test_choice_equality(self):
        """Test choice equality."""
        choice1 = Choice("Same text")
        choice2 = Choice("Same text")
        choice3 = Choice("Different text")

        assert choice1 == choice2
        assert choice1 != choice3


class TestEvent:
    """Test cases for Event model."""

    def test_event_creation_minimal(self):
        """Test minimal event creation."""
        event = Event(id="test", description="Test event")
        assert event.id == "test"
        assert event.description == "Test event"
        assert event.choices == []
        assert event.items == []
        assert event.location == ""
        assert event.act == 0
        assert event.scene == 0

    def test_event_creation_full(self, sample_choice):
        """Test full event creation."""
        choices = [sample_choice]
        items = ["sword", "potion"]
        event = Event(
            id="full_event",
            description="Full test event",
            choices=choices,
            items=items,
            location="Test location",
            act=1,
            scene=2,
        )

        assert event.id == "full_event"
        assert event.description == "Full test event"
        assert event.choices == choices
        assert event.items == items
        assert event.location == "Test location"
        assert event.act == 1
        assert event.scene == 2

    def test_event_choices_modification(self):
        """Test that event choices can be modified."""
        event = Event(id="test", description="Test")
        choice = Choice("New choice")
        event.choices.append(choice)
        assert len(event.choices) == 1
        assert event.choices[0] == choice


class TestGameState:
    """Test cases for GameState model."""

    def test_game_state_creation_default(self):
        """Test default game state creation."""
        state = GameState()
        assert state.current_event is None
        assert state.inventory == []
        assert state.act == 0
        assert state.scene == 0
        assert state.visited_events == []
        assert state.flags == {}

    def test_game_state_creation_with_values(self, sample_event):
        """Test game state creation with values."""
        inventory = ["sword", "potion"]
        visited = ["event_1", "event_2"]
        flags = {"started": True, "level": 5}

        state = GameState(
            current_event=sample_event,
            inventory=inventory,
            act=2,
            scene=3,
            visited_events=visited,
            flags=flags,
        )

        assert state.current_event == sample_event
        assert state.inventory == inventory
        assert state.act == 2
        assert state.scene == 3
        assert state.visited_events == visited
        assert state.flags == flags

    def test_game_state_modification(self):
        """Test that game state can be modified."""
        state = GameState()
        state.inventory.append("new_item")
        state.flags["new_flag"] = True

        assert "new_item" in state.inventory
        assert state.flags["new_flag"] is True


class TestActionLogEntry:
    """Test cases for ActionLogEntry model."""

    def test_action_log_entry_creation(self):
        """Test action log entry creation."""
        entry = ActionLogEntry(
            timestamp=123, action_type="test_action", details="Test details"
        )

        assert entry.timestamp == 123
        assert entry.action_type == "test_action"
        assert entry.details == "Test details"

    def test_action_log_entry_equality(self):
        """Test action log entry equality."""
        entry1 = ActionLogEntry(1, "action", "details")
        entry2 = ActionLogEntry(1, "action", "details")
        entry3 = ActionLogEntry(2, "action", "details")

        assert entry1 == entry2
        assert entry1 != entry3


class TestStoryData:
    """Test cases for StoryData model."""

    def test_story_data_creation_default(self):
        """Test default story data creation."""
        story = StoryData()
        assert story.title == "Text Adventure"
        assert story.setting == "Unknown"
        assert story.outline == ""
        assert story.acts == []
        assert story.npcs == []
        assert story.items == {}

    def test_story_data_creation_with_values(self):
        """Test story data creation with values."""
        acts = [{"name": "Act 1", "description": "First act"}]
        npcs = [{"name": "Hero", "description": "Main character"}]
        items = {"sword": {"name": "Sword", "description": "Sharp blade"}}

        story = StoryData(
            title="Test Story",
            setting="Fantasy World",
            outline="Epic adventure",
            acts=acts,
            npcs=npcs,
            items=items,
        )

        assert story.title == "Test Story"
        assert story.setting == "Fantasy World"
        assert story.outline == "Epic adventure"
        assert story.acts == acts
        assert story.npcs == npcs
        assert story.items == items

    def test_story_data_modification(self):
        """Test that story data can be modified."""
        story = StoryData()
        story.acts.append({"name": "New Act"})
        story.npcs.append({"name": "New NPC"})
        story.items["new_item"] = {"name": "New Item"}

        assert len(story.acts) == 1
        assert len(story.npcs) == 1
        assert "new_item" in story.items
